from tfinterpy.idw import IDW
from tfinterpy.grid import Grid3D
from tfinterpy.gslib.fileUtils import readGslibPoints
import tfinterpy.vtk.colorMap as CM
from tfinterpy.vtk.rendering import createGridActor, rendering
from tfinterpy.vtk.fileUtils import saveVTKGrid

if __name__ == "__main__":
    filePath = "data/sample_data.gslib"
    N = 8
    df = readGslibPoints(filePath)# Read 3d points from gslib file.
    print(df.describe())
    samples = df[['x', 'y', 'z', 'porosity']].values

    # Create linear 3D grid.
    grid = Grid3D()
    grid.rectlinear((100, 100, 10), (samples[:, 0].min(), samples[:, 0].max()),
                    (samples[:, 1].min(), samples[:, 1].max()), (samples[:, 2].min(), samples[:, 2].max()))
    exe = IDW(samples, '3d')# Create a idw interpolator.

    grid.pro = exe.execute(grid.points(), N)# Perform interpolation of all points in the grid.
    print(exe.crossValidate(N))# Perform leave-one-out validation and print result.
    print(exe.crossValidateKFold(10, N))# Perform k-fold validation and print result.

    # Create an actor representing a rectilinear grid and use the interpolation results for color mapping.
    actor = createGridActor(*grid.dim, grid.x, grid.y, grid.z, grid.pro,
                            [samples[:, 3].min(), samples[:, 3].max()], CM.Rainbow)
    rendering(actor)# Rendering vtkActor.
    saveVTKGrid('savedData/grid.vtk', actor.GetMapper().GetInput())# Save grid data to vtk file.



import numpy as np

EPS = 1e-16

def search2d(vecs, vars, lagNum, lagInterval, lagTole, angle, angleTole, bandWidth):
    '''
    Search for lags in two dimensions.

    :param vecs: ndarray, vectors.
    :param vars: ndarray, semivariance.
    :param lagNum: integer, number of lags to search.
    :param lagInterval: number, interval lags' distance.
    :param lagTole: number, tolerance of lag's distance.
    :param angle: number, search direction, angle range is [0, PI].
    :param angleTole: number, angle tolerance.
    :param bandWidth: number, bandwidth used during search.
    :return: tuple, (processedLags, lags).
        processedLags is the mean-processed lags.
        lags is original searched lags.
    '''
    unitVector = np.array([np.cos(angle), np.sin(angle)])
    return searchLags(vecs, vars, lagNum, lagInterval, lagTole, unitVector, angleTole, bandWidth)


def search3d(vecs, vars, lagNum, lagInterval, lagTole, azimuth, dip, angleTole, bandWidth):
    '''

    :param vecs: ndarray, vectors.
    :param vars: ndarray, semivariance.
    :param lagNum: integer, number of lags to search.
    :param lagInterval: number, interval lags' distance.
    :param lagTole: number, tolerance of lag's distance.
    :param azimuth: number, azimuth of search direction, the range of azimuth is [0, PI].
    :param dip: number, dip of search direction, the range of dip is [0, PI].
    :param angleTole: number, angle tolerance.
    :param bandWidth: number, bandwidth used during search.
    :return: tuple, (processedLags, lags).
        processedLags is the mean-processed lags.
        lags is original searched lags.
    '''
    unitVector = np.array([np.cos(azimuth) * np.cos(dip), np.sin(azimuth) * np.cos(dip), np.sin(dip)])
    return searchLags(vecs, vars, lagNum, lagInterval, lagTole, unitVector, angleTole, bandWidth)


def searchLags(vecs, vars, lagNum, lagInterval, lagTole, unitVector, angleTole, bandWidth):
    '''
    Search lags.

    :param vecs: ndarray, vectors.
    :param vars: ndarray, semivariance.
    :param lagNum: integer, number of lags to search.
    :param lagInterval: number, interval lags' distance.
    :param lagTole: number, tolerance of lag's distance.
    :param unitVector: array_like, unit vector of search direction.
    :param angleTole: number, angle tolerance.
    :param bandWidth: number, bandwidth used during search.
    :return: tuple, (processedLags, lags).
        processedLags is the mean-processed lags.
        lags is original searched lags.
    '''
    norms = np.linalg.norm(vecs, axis=1) + EPS
    thetas = np.arccos(np.dot(vecs, unitVector) / norms)
    indice = np.where(thetas > np.pi / 2)[0]
    thetas[indice] = np.pi - thetas[indice]
    # filter with angle
    remain = np.where(thetas <= angleTole)[0]
    norms = norms[remain]
    thetas = thetas[remain]
    vars = vars[remain]
    # filter with bandwith
    bands = norms * np.sin(thetas)
    remain = np.where(bands <= bandWidth)[0]
    norms = norms[remain]
    vars = vars[remain]
    # filter with lag range
    lagRanList = []
    for i in range(1, lagNum + 1):
        lagRanList.append([lagInterval * i - lagTole, lagInterval * i + lagTole])
    lags = [[] for i in range(lagNum)]
    for i in range(len(norms)):
        h = norms[i]
        for j in range(len(lagRanList)):
            if h > lagRanList[j][0] and h < lagRanList[j][1]:
                lags[j].append([h, vars[i]])
                break
    processedLags = []
    for ls in lags:
        if len(ls)==0:
            continue
        mean = np.mean(ls, axis=0)
        if np.isnan(mean).any():
            continue
        processedLags.append(mean)
    return processedLags, lags



import numpy as np


class Grid2D:
    '''
    Classes used to represent a two-dimensional grid.
    '''

    def __init__(self):
        self.x = None
        self.y = None
        self.xran = None
        self.yran = None
        self._xy = None
        self.dim = [-1, -1]  #width and height

    def rectlinear(self, dim, xran, yran):
        '''
        Initialize the two-dimensional grid according to the dimension, x range, y range.

        :param dim: array_like, array containing two integer. [width, height].
        :param xran: array_like, [xbegin, xend]. So the step of x is equal to (xend-xbegin)/(width-1).
        :param yran: array_like, [ybegin, yend].
        :return: None
        '''
        self.dim = dim
        self.xran = xran
        self.yran = yran
        xstep = (xran[1] - xran[0]) / (dim[0] - 1)
        self.x = np.array([(i * xstep + xran[0]) for i in range(dim[0])])
        ystep = (yran[1] - yran[0]) / (dim[1] - 1)
        self.y = np.array([(i * ystep + yran[0]) for i in range(dim[1])])
        self._xy = None

    def setXCoords(self, xcoords):
        '''
        Sets the x coordinate of the grid.

        :param xcoords: array_like, array containing all the coordinates along the x axis.
        :return: None
        '''
        self.x = np.array(xcoords)
        self.dim[0] = len(self.x)
        self.xran = [self.x.min(), self.x.max()]
        self._xy = None

    def setYCoords(self, ycoords):
        '''
        Sets the y coordinate of the grid.

        :param ycoords: array_like, array containing all the coordinates along the y axis.
        :return: None
        '''
        self.y = np.array(ycoords)
        self.dim[1] = len(self.y)
        self.yran = [self.y.min(), self.y.max()]
        self._xy = None

    def points(self):
        '''
        Returns an ndarray containing all coordinates on the grid,
        and the shape of the array is (width*height, 2).

        :return: ndarray
        '''
        if self._xy is None:
            self._xy = np.zeros((self.dim[1], self.dim[0], 2))
            for i in range(self.dim[1]):
                self._xy[i, :, 0] = self.x
            for i in range(self.dim[0]):
                self._xy[:, i, 1] = self.y
            self._xy.resize((self.dim[0] * self.dim[1], 2))
        return self._xy


class Grid3D:
    '''
    Classes used to represent a three-dimensional grid.
    '''

    def __init__(self):
        self.x = None
        self.y = None
        self.z = None
        self.xran = None
        self.yran = None
        self.zran = None
        self._xyz = None
        self.dim = [-1, -1, -1] # width(x), height(y), long(z)

    def rectlinear(self, dim, xran, yran, zran):
        '''
        Initialize the three-dimensional grid according to the dimension, x range, y range, xrange.

        :param dim: array_like, array containing two integer. [width, height].
        :param xran: array_like, [xbegin, xend]. So the step of x is equal to (xend-xbegin)/(width-1).
        :param yran: array_like, [ybegin, yend].
        :param zran: array_like, [zbegin, zend].
        :return: None
        '''
        self.dim = dim
        self.xran = xran
        self.yran = yran
        self.zran = zran
        xstep = (xran[1] - xran[0]) / (dim[0] - 1)
        self.x = np.array([(i * xstep + xran[0]) for i in range(dim[0])])
        ystep = (yran[1] - yran[0]) / (dim[1] - 1)
        self.y = np.array([(i * ystep + yran[0]) for i in range(dim[1])])
        zstep = (zran[1] - zran[0]) / (dim[2] - 1)
        self.z = np.array([(i * zstep + zran[0]) for i in range(dim[2])])
        self._xyz = None

    def setXCoords(self, xcoords):
        '''
        Sets the x coordinate of the grid.

        :param xcoords: array_like, array containing all the coordinates along the x axis.
        :return: None
        '''
        self.x = np.array(xcoords)
        self.dim[0] = len(self.x)
        self.xran = [self.x.min(), self.x.max()]
        self._xyz = None

    def setYCoords(self, ycoords):
        '''
        Sets the y coordinate of the grid.

        :param ycoords: array_like, array containing all the coordinates along the y axis.
        :return: None
        '''
        self.y = np.array(ycoords)
        self.dim[1] = len(self.y)
        self.yran = [self.y.min(), self.y.max()]
        self._xyz = None

    def setZCoords(self, zcoords):
        '''
        Sets the z coordinate of the grid.

        :param zcoords: array_like, array containing all the coordinates along the z axis.
        :return: None
        '''
        self.z = np.array(zcoords)
        self.dim[2] = len(self.z)
        self.zran = [self.z.min(), self.z.max()]
        self._xyz = None

    def points(self):
        '''
        Returns an ndarray containing all coordinates on the grid,
        and the shape of the array is (width*height*long, 3).

        :return: ndarray
        '''
        if self._xyz is None:
            self._xyz = np.zeros((self.dim[2], self.dim[1], self.dim[0], 3))
            for i in range(self.dim[2]):
                for j in range(self.dim[1]):
                    self._xyz[i, j, :, 0] = self.x
                for j in range(self.dim[0]):
                    self._xyz[i, :, j, 1] = self.y
                self._xyz[i, :, :, 2] = self.z[i]
            self._xyz.resize((self.dim[0] * self.dim[1] * self.dim[2], 3))
        return self._xyz



from tfinterpy.krige import SK
from tfinterpy.grid import Grid3D
from tfinterpy.variogram import calculateDefaultVariogram3D
from tfinterpy.gslib.fileUtils import readGslibPoints
from tfinterpy.vtk.rendering import createGridActor, rendering
import tfinterpy.vtk.colorMap as CM
from tfinterpy.vtk.fileUtils import saveVTKGrid
import matplotlib.pyplot as plt

if __name__ == "__main__":
    filePath = "data/sample_data.gslib"
    N = 8
    df = readGslibPoints(filePath)
    print(df.describe())
    samples = df[['x', 'y', 'z', 'porosity']].values

    # Create linear 3D grid.
    grid = Grid3D()
    grid.rectlinear((100, 100, 10), (samples[:, 0].min(), samples[:, 0].max()),
                    (samples[:, 1].min(), samples[:, 1].max()), (samples[:, 2].min(), samples[:, 2].max()))
    exe = SK(samples, '3d')# Create a sk interpolator.
    vb = calculateDefaultVariogram3D(samples)# Calculate a default variogram function.
    plt.figure()
    vb.showVariogram()
    plt.show()

    # Perform interpolation of all points in the grid.
    grid.pro, grid.sigma = exe.execute(grid.points(), N, vb.getVariogram())

    print(exe.crossValidate(N, vb.getVariogram()))# Perform leave-one-out validation and print result.
    print(exe.crossValidateKFold(10, N, vb.getVariogram()))# Perform k-fold validation and print result.

    # Create an actor representing a rectilinear grid and use the interpolation results for color mapping.
    actor = createGridActor(*grid.dim, grid.x, grid.y, grid.z, grid.pro, [samples[:, 3].min(), samples[:, 3].max()])
    rendering(actor)# Rendering vtkActor.
    # Create an actor representing a rectilinear grid and use the kriging variances for color mapping.
    actor = createGridActor(*grid.dim, grid.x, grid.y, grid.z, grid.sigma, None, CM.Rainbow)
    rendering(actor)# Rendering vtkActor.
    saveVTKGrid('savedData/grid.vtk', actor.GetMapper().GetInput())# Save grid data(kriging variances) to vtk file.



# -*- coding: utf-8 -*-
# @Time    : 2022/3/28 12:08
# @Author  : UCCU
import vtk
import tfinterpy.vtk.colorMap as CM
import numpy as np

def createPointsActor(x,y,z=0,scalar=None,scalarRange=None,colorMap=CM.BlueGreenOrange):
    '''
    Create vertices representing coordinate points based on the coordinate data.

    :param x: array_like, x coordinate data.
    :param y: array_like, y coordinate data.
    :param z: array_like, z coordinate data.
        If it is a two-dimensional space, z should be set to 0.
    :param scalar: array_like, properties.
    :param scalarRange: array_like, [scalarMin, scalarMax], properties' range.
    :param colorMap: list, refer to the colorMap in the vtk.colorMap module.
    :return: vtkActor object.
    '''
    points = vtk.vtkPoints()
    L=len(x)
    points.SetNumberOfPoints(L)
    if z==0:
        for i in range(L):
            points.SetPoint(i,x[i],y[i],0)
    else:
        for i in range(L):
            points.SetPoint(i,x[i],y[i],z[i])
    polyData = vtk.vtkPolyData()
    polyData.SetPoints(points)
    scalarRange=None
    if scalar is not None:
        scalarArr = vtk.vtkFloatArray()
        scalarArr.SetName("scalar")
        scalarArr.SetNumberOfComponents(1)
        scalarArr.SetNumberOfValues(L)
        for i in range(L):
            scalarArr.SetTuple1(i,scalar[i])
        polyData.GetPointData().AddArray(scalarArr)
        polyData.GetPointData().SetActiveScalars("scalar")
        if scalarRange is None:
            scalarRange=[scalar.min(),scalar.max()]
    vertexGlyphFilter = vtk.vtkVertexGlyphFilter()
    vertexGlyphFilter.AddInputData(polyData)
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(vertexGlyphFilter.GetOutputPort())
    if scalarRange is not None:
        mapper.SetScalarRange(*scalarRange)
        ctf = CM.getCTF(scalarRange[0], scalarRange[1], colorMap)
        ctf.SetDiscretize(True)
        ctf.Build()
        mapper.SetLookupTable(ctf)
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetPointSize(5)
    return actor

def createGridActor(nx,ny,nz,x,y,z,scalar=None,scalarRange=None,colorMap=CM.Rainbow):
    '''
    Create rectilinear grid based on the coordinate data.

    :param nx: integer, number of x coordinate.
    :param ny: integer, number of y coordinate.
    :param nz: integer, number of z coordinate.
    :param x: array_like, x coordinate data.
    :param y: array_like, y coordinate data.
    :param z: array_like, z coordinate data.
    :param scalar: array_like, properties.
    :param scalarRange: array_like, [scalarMin, scalarMax], properties' range.
    :param colorMap: list, refer to the colorMap in the vtk.colorMap module.
    :return: vtkActor object.
    '''
    rgrid = vtk.vtkRectilinearGrid()
    nz=1 if nz<=1 else nz
    z=[0] if nz==1 else z
    dim=[nx, ny, nz]
    rgrid.SetDimensions(dim)
    xcoords = vtk.vtkFloatArray()
    xcoords.SetNumberOfValues(dim[0])
    xcoords.SetNumberOfComponents(1)
    ycoords = vtk.vtkFloatArray()
    ycoords.SetNumberOfValues(dim[1])
    ycoords.SetNumberOfComponents(1)
    zcoords = vtk.vtkFloatArray()
    zcoords.SetNumberOfValues(dim[2])
    zcoords.SetNumberOfComponents(1)
    for idx, value in enumerate(x):
        xcoords.SetValue(idx, value)
    for idx, value in enumerate(y):
        ycoords.SetValue(idx, value)
    for idx, value in enumerate(z):
        zcoords.SetValue(idx, value)
    rgrid.SetXCoordinates(xcoords)
    rgrid.SetYCoordinates(ycoords)
    rgrid.SetZCoordinates(zcoords)
    if scalar is not None:
        scalarArr = vtk.vtkFloatArray()
        scalarArr.SetName('scalar')
        scalarArr.SetNumberOfComponents(1)
        scalarArr.SetNumberOfValues(len(scalar))
        for idx,value in enumerate(scalar):
            scalarArr.SetValue(idx,value)
        rgrid.GetPointData().AddArray(scalarArr)
        rgrid.GetPointData().SetActiveScalars("scalar")
        if scalarRange is None:
            scalarRange = [np.percentile(scalar,1), np.percentile(scalar,99)]
    mapper = vtk.vtkDataSetMapper()
    mapper.SetInputData(rgrid)
    if scalarRange is not None:
        mapper.SetScalarRange(*scalarRange)
        ctf = CM.getCTF(scalarRange[0], scalarRange[1], colorMap)
        ctf.SetDiscretize(True)
        ctf.Build()
        mapper.SetLookupTable(ctf)
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetInterpolationToGouraud()
    return actor

def rendering(actor):
    '''
    Create a new render window to render the actor.

    :param actor: vtkActor object.
    :return: None.
    '''
    ren = vtk.vtkRenderer()
    ren.AddActor(actor)
    renWin = vtk.vtkRenderWindow()
    renWin.AddRenderer(ren)
    iren = vtk.vtkRenderWindowInteractor()
    iren.SetRenderWindow(renWin)
    ren.SetBackground(0.447, 0.552, 0.756)
    renWin.SetSize(960, 680)
    iren.Initialize()
    ren.ResetCamera()
    renWin.Render()
    iren.Start()


from tfinterpy.krige import OK
from tfinterpy.grid import Grid3D
from tfinterpy.variogram import calculateOmnidirectionalVariogram3D
from tfinterpy.gslib.fileUtils import readGslibPoints
from tfinterpy.vtk.rendering import createGridActor, rendering
import tfinterpy.vtk.colorMap as CM
from tfinterpy.vtk.fileUtils import saveVTKGrid
from tfinterpy.utils import calcVecs
import matplotlib.pyplot as plt
import time

if __name__ == "__main__":
    filePath = "data/sample_data.gslib"
    N = 8
    df = readGslibPoints(filePath)# Read 3d points from gslib file.
    print(df.describe())
    samples = df[['x', 'y', 'z', 'porosity']].values

    # Create linear 3D grid.
    grid = Grid3D()
    grid.rectlinear((100, 100, 10), (samples[:, 0].min(), samples[:, 0].max()),
                    (samples[:, 1].min(), samples[:, 1].max()), (samples[:, 2].min(), samples[:, 2].max()))

    # Calculate a nested variogram function.
    nestVariogram, variogramBuilders = calculateOmnidirectionalVariogram3D(samples)
    for vb in variogramBuilders:
        plt.figure()
        vb.showVariogram()
        plt.show()

    exe = OK(samples, '3d')# Create a ok interpolator.
    t1=time.perf_counter()
    # Perform interpolation of all points in the grid.
    grid.pro, grid.sigma = exe.execute(grid.points(), N, nestVariogram, workerNum=3)
    print('predict time,',time.perf_counter()-t1)

    print(exe.crossValidate(N, nestVariogram))# Perform leave-one-out validation and print result.
    print(exe.crossValidateKFold(10, N, nestVariogram))# Perform k-fold validation and print result.
    # Create an actor representing a rectilinear grid and use the interpolation results for color mapping.
    actor = createGridActor(*grid.dim, grid.x, grid.y, grid.z, grid.pro,
                            [samples[:, 3].min(), samples[:, 3].max()], CM.Rainbow)
    rendering(actor)# Rendering vtkActor.
    # Create an actor representing a rectilinear grid and use the kriging variances for color mapping.
    actor = createGridActor(*grid.dim, grid.x, grid.y, grid.z, grid.sigma, None, CM.Rainbow)
    rendering(actor)# Rendering vtkActor.
    saveVTKGrid('savedData/grid.vtk', actor.GetMapper().GetInput())# Save grid data(kriging variances) to vtk file.



import numpy as np


def h2(p1, p2):
    '''
    Return the square of distance between p1 and p2.
    p1 and p2 can be either a one-dimensional array representing a single
    coordinate point or an ndarray containing multiple coordinate points.
    The shapes of p1 and p2 should be the same.

    :param p1: array_like.
    :param p2: array_like.
    :return: ndarray.
    '''
    diff = np.square(np.subtract(p1, p2))
    return np.sum(diff, axis=-1)


def h(p1, p2):
    '''
    Return the distance between p1 and p2.
    p1 and p2 can be either a one-dimensional array representing a single
    coordinate point or an ndarray containing multiple coordinate points.
    The shapes of p1 and p2 should be the same.

    :param p1: array_like.
    :param p2: array_like.
    :return: ndarray.
    '''
    return np.sqrt(h2(p1, p2))


def ha(p1, p2):
    '''
    Return the distance and angele between p1 and p2.
    p1 and p2 can be either a one-dimensional array representing a single
    coordinate point or an ndarray containing multiple coordinate points.
    The shapes of p1 and p2 should be the same.

    :param p1: array_like.
    :param p2: array_like.
    :return: tuple, tuple containing two ndarray.
        The first ndarray representing the distance,
        the second list representing the angle, the angle range is [0,PI].
    '''
    diff = np.subtract(p1, p2)
    h = np.sqrt(np.sum(np.square(diff), axis=-1))
    if diff.ndim == 1:
        angle = np.arctan2(diff[1], diff[0])
        angle = angle + np.pi if angle < 0 else angle
    else:
        angle = np.arctan2(diff[:, 1], diff[:, 0])
        angle[angle < 0] += np.pi
    return h, angle


def calcHV(data, dim=2, attrib_col=2):
    '''
    Calculates the distance and variation between data.

    :param data: ndarray, array containing coordinates and attributes.
    :param dim: integer, indicates whether the coordinates are 2d or 3d.
    :param attrib_col: integer, indicates which column is property.
    :return: ndarray, the first column is distance, the second column is variation.
    '''
    if dim == 3 and attrib_col < 3:
        attrib_col = 3
    arr = np.empty((0, 2), dtype="float32")
    size = data.shape[0]
    for i in range(size - 1):
        dis = h(data[i + 1:, :dim], data[i, :dim])
        halfvar = 0.5 * ((data[i + 1:, attrib_col] - data[i, attrib_col]) ** 2)
        dis.resize((dis.shape[0], 1))
        halfvar.resize((halfvar.shape[0], 1))
        arr = np.append(arr, np.concatenate([dis, halfvar], axis=1), axis=0)
    return arr


def calcHAV(data, attrib_col=2):
    '''
    Calculates the distance, angle and variation between data.

    :param data: ndarray, array containing coordinates and attributes.
    :param attrib_col: integer, indicates which column is property.
    :return: ndarray, the first column is distance,
        the second column is angle (range is [0, PI]),
        the third column is variation.
    '''
    dim = 2
    if attrib_col < 2:
        attrib_col = 2
    arr = np.empty((0, 3), dtype="float32")
    size = data.shape[0]
    for i in range(size - 1):
        dis, angle = ha(data[i + 1:, :dim], data[i, :dim])
        halfvar = 0.5 * ((data[i + 1:, attrib_col] - data[i, attrib_col]) ** 2)
        dis.resize((dis.shape[0], 1))
        angle.resize((angle.shape[0], 1))
        halfvar.resize((halfvar.shape[0], 1))
        arr = np.append(arr, np.concatenate([dis, angle, halfvar], axis=1), axis=0)
    return arr


def calcHAVByVecs(vecs):
    '''
    Calculates the distance, angle and variation according to vectors.

    :param vecs: ndarray, vectors containing 2d coordinates' difference and properties' difference.
    :return: ndarray, the first column is distance,
        the second column is angle (range is [0, PI]),
        the third column is variation.
    '''
    distance = np.linalg.norm(vecs[:, :2], axis=1)
    angle = np.arctan2(vecs[:, 1], vecs[:, 0])
    angle[angle < 0] += np.pi
    var = 0.5 * (vecs[:, 2] ** 2)
    hav = np.zeros((len(vecs), 3))
    hav[:, 0] = distance
    hav[:, 1] = angle
    hav[:, 2] = var
    return hav


def calcHABVByVecs(vecs):
    '''
    Calculates the distance, azimuth(alpha), dip(beta) and variation according to vectors.

    :param vecs: ndarray, vectors containing 3d coordinates' difference and properties' difference.
    :return: ndarray, the first column is distance,
        the second column is azimuth(alpha) (range is [0, PI]),
        the second column is dip(beta) (range is [0, PI]),
        the fourth column is variation.
    '''
    distance = np.linalg.norm(vecs[:, :3], axis=1)
    azimuth = np.arctan2(vecs[:, 1], vecs[:, 0])
    azimuth[azimuth < 0] += np.pi
    norm = np.linalg.norm(vecs[:, :2], axis=1)
    dip = np.arctan2(vecs[:, 2], norm)
    dip[dip < 0] += np.pi
    var = 0.5 * (vecs[:, 3] ** 2)
    habv = np.zeros((len(vecs), 4))
    habv[:, 0] = distance
    habv[:, 1] = azimuth
    habv[:, 2] = dip
    habv[:, 3] = var
    return habv


def calcVecs(data, includeSelf=False, repeat=True):
    '''
    Compute vectors between data.

    :param data: ndarray, two-dimensional array.
    :param includeSelf: boolean, whether to subtract the vector from itself.
    :param repeat: boolean, whether to calculate the two points that have subtracted from each other.
    :return: ndarray, vectors.
    '''
    L = len(data)
    if repeat:
        vecs = []
        if includeSelf:
            for i in range(L):
                vec = data - data[i]
                vecs.append(vec)
        else:
            _indice = [i for i in range(L)]
            for i in range(L):
                indice = list(_indice)
                indice.pop(i)
                vec = data[indice] - data[i]
                vecs.append(vec)
        vecs = np.array(vecs)
        return vecs.reshape((vecs.shape[0] * vecs.shape[1], vecs.shape[2]))
    else:
        vecs = np.empty((0, data.shape[1]))
        for i in range(L - 1):
            vec = data[i + 1:] - data[i]
            vecs = np.append(vecs, vec, axis=0)
        return vecs


def kSplit(ndarray, k):
    '''
    Split the ndarray into k fold.

    :param ndarray: ndarray.
    :param k: integer, fold number.
    :return: list, list containing all fold.
    '''
    size = ndarray.shape[0]
    step = size // k
    if size % k != 0:
        step += 1
    splits = []
    for i in range(k):
        begin = int(i * step)
        end = int((i + 1) * step)
        if i == k - 1:
            splits.append(ndarray[begin:])
        else:
            splits.append(ndarray[begin:end])
    return splits



import numpy as np


def linear(slope, C0, h):
    '''
    Linear model.

    :param slope: number.
    :param C0: number, nugget.
    :param h: number or ndarray, distance.
    :return: number or ndarray.
    '''
    return slope * h + C0


def power(scale, exp, C0, h):
    '''
    Power model.

    :param scale: number.
    :param exp: number.
    :param C0: number, nugget.
    :param h: number or ndarray, distance.
    :return: number or ndarray.
    '''
    return scale * (h ** exp) + C0


def gaussian(C, a, C0, h):
    '''
    Gaussian model.

    :param C: number, sill.
    :param a: number, range.
    :param C0: number, nugget.
    :param h: number or ndarray, distance.
    :return: number or ndarray.
    '''
    return C * (1.0 - np.exp(-(h ** 2.0) / (a * 4.0 / 7.0) ** 2.0)) + C0


def exponent(C, a, C0, h):
    '''
    Exponent model.

    :param C: number, sill.
    :param a: number, range.
    :param C0: number, nugget.
    :param h: number or ndarray, distance.
    :return: number or ndarray.
    '''
    return C * (1.0 - np.exp(-h / (a / 3.0))) + C0


def spherical(C, a, C0, h):
    '''
    Spherical model.

    :param C: number, sill.
    :param a: number, range.
    :param C0: number, nugget.
    :param h: number or ndarray, distance.
    :return: number or ndarray.
    '''
    return np.piecewise(h, [h <= a, h > a],
                        [lambda x: C * ((3.0 * x) / (2.0 * a) - (x ** 3.0) / (2.0 * a ** 3.0)) + C0, C + C0, ])


VariogramModelMap = {
    "spherical": spherical,
    "gaussian": gaussian,
    "exponent": exponent,
    "linear": linear,
    "power": power,
}


def variogramModel(name):
    '''
    Returns the variogram function based on the name.

    :param name: str, variogram function name.
    :return: function.
    '''
    func = VariogramModelMap.get(name)
    return func



import netCDF4 as nc
import random
import numpy as np
import os
import os.path as op


def getNCFileList(dirPath):
    '''
    Returns all netcdf files' path under the specified directory.

    :param dirPath: str, directory.
    :return: list, files' path.
    '''
    ls = []
    if not op.exists(dirPath):
        return ls
    for file in os.listdir(dirPath):
        if file[-3:] == '.nc':
            ls.append(file)
    return ls


def peakNCWH(filePath, wk="lon", hk="lat"):
    '''
    Get the width and height information contained in the nc file.

    :param filePath: str.
    :param wk: str, the key corresponding to the width.
    :param hk: str, the key corresponding to the height.
    :return: tuple, (width, height).
        If the file does not contain width and height information,
        return (None,None).
    '''
    if not op.exists(filePath):
        return None, None
    try:
        dem = nc.Dataset(filePath)
    except:
        print("netcdf 文件读取异常")
        return None, None
    W = dem.dimensions[wk].size
    H = dem.dimensions[hk].size
    dem.close()
    return W, H


def peakNCInfo(filePath):
    '''
     Get the key and corresponding shape information contained in the nc file.

    :param filePath: str.
    :return: tuple, (keys,info).
        keys is list containing all key in nc file.
        info is a string that records the shape of the variables.
    '''
    if not op.exists(filePath):
        return None, None
    try:
        dem = nc.Dataset(filePath)
    except:
        print("netcdf 文件读取异常")
        return None, None
    info = ''
    keys = []
    for key in dem.variables.keys():
        if dem.variables[key].shape == ():
            continue
        keys.append(key)
        info += "%s shape: %s\n" % (key, str(dem.variables[key].shape))
    dem.close()
    return keys, info


def getSamples(filePathOrData, X, Y, Z, begin=(0, 0), sub=(200, 200), N=10):
    '''
    :param filePath: str.
    :param X: str, key of lats.
    :param Y: str, key of lons.
    :param Z: str, key of ele.
    :param begin: tuple, the begining index of the submap.
    :param sub: tuple, the size of the submap.
    :param N: integer, number of samples in X and Y directions.
    :return: tuple, (samples,lats,lons,ele).
        samples is the sampling point obtained by random sampling within the submap.
        lats is latitude of submap.
        lons is longitude of submap.
        ele is elevation of submap.
    '''
    if type(filePathOrData) == type(''):
        dem = nc.Dataset(filePathOrData)
        data = dem.variables
    else:
        data = filePathOrData
    by, bx = begin
    suby, subx = sub
    lats = data[X][bx:bx + subx]  # 纬度值
    lons = data[Y][by:by + suby]  # 经度值
    ele = data[Z][bx:bx + subx, by:by + suby]  # 海拔值
    if type(filePathOrData) == type(''):
        dem.close()  #
    subx, suby = sub
    samples = []
    c = 0
    while c < N:
        x = random.randint(0, subx - 1)
        y = random.randint(0, suby - 1)
        if [x, y, ele[y, x]] not in samples:
            samples.append([x, y, ele[y, x]])
            c += 1
    samples = np.array(samples, dtype="float32")
    return samples, lats, lons, ele



from tfinterpy.gui.uiTool import *
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
import sys
import numpy as np
import pandas as pd

from tfinterpy.variogramModel import VariogramModelMap
from tfinterpy.variogramExp import search2d, search3d
from tfinterpy.variogram import VariogramBuilder, calculateDefaultVariogram2D, calculateDefaultVariogram3D, \
    calculateOmnidirectionalVariogram2D, calculateOmnidirectionalVariogram3D
from tfinterpy.utils import calcVecs, calcHAVByVecs, calcHABVByVecs, calcHV
from tfinterpy.gui.canvas import Canvas
from tfinterpy.grid import Grid2D, Grid3D
from tfinterpy.vtk import colorMap
import vtk
from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
import tensorflow as tf
from tfinterpy.idw import IDW
from tfinterpy.krige import SK, OK
from tfinterpy.tf.idw import TFIDW
from tfinterpy.tf.krige import TFSK, TFOK
from tfinterpy.tf.variogramLayer import getVariogramLayer, getNestVariogramLayer
from tfinterpy.gslib.fileUtils import readGslibPoints, saveGslibPoints, saveGslibGrid
from tfinterpy.vtk.fileUtils import saveVTKPoints, saveVTKGrid
from scipy.optimize import least_squares
import time

ColorMap = colorMap.Rainbow #default color map


class Tool(QMainWindow, Ui_MainWindow):
    def __init__(self):
        super(Tool, self).__init__()
        self.setupUi(self)
        self.actorPS = None
        self.actorGrid = None
        self.actorAxes = None
        self.actorScalarBar = None
        self.initUi()
        self.connectSignals()
        self.data = None
        self.samples = None
        self.vl = None
        self.vb = None
        self.grid = None


    def initUi(self):
        self.vtkWidget = QVTKRenderWindowInteractor(self.tab_interp)
        self.ren = vtk.vtkRenderer()
        self.vtkWidget.GetRenderWindow().AddRenderer(self.ren)
        self.iren = self.vtkWidget.GetRenderWindow().GetInteractor()
        self.actorScalarBar = vtk.vtkScalarBarActor()
        self.actorScalarBar.GetPositionCoordinate().SetCoordinateSystemToNormalizedViewport()
        self.actorScalarBar.GetPositionCoordinate().SetValue(0.88, 0.3)
        self.actorScalarBar.SetWidth(0.1)
        self.actorScalarBar.SetHeight(0.5)
        self.actorScalarBar.SetOrientationToVertical()
        self.actorScalarBar.SetNumberOfLabels(8)
        self.actorScalarBar.SetTextPositionToSucceedScalarBar()
        self.actorScalarBar.SetVisibility(True)
        self.ren.SetBackground(0.447, 0.552, 0.756)

        self.iren.Start()

        vl = QVBoxLayout(self.tab_interp)
        vl.addWidget(self.vtkWidget)
        self.tab_interp.setLayout(vl)

        self.canvasVariogram = Canvas(1, 1, 100, self.tab_variogram)
        # self.axesVariogram = self.canvasVariogram.fig.add_subplot(111)
        self.vl_variogram.addWidget(self.canvasVariogram)

        self.canvasCV = Canvas(1, 1, 100, self.tab_cv)
        self.axesCV = self.canvasCV.fig.add_subplot(111)
        vl = QVBoxLayout(self.tab_cv)
        vl.addWidget(self.canvasCV)
        self.tab_cv.setLayout(vl)

        self.le_alpha.setValidator(QDoubleValidator())
        self.le_angle.setValidator(QDoubleValidator())
        self.le_angleDip.setValidator(QDoubleValidator())
        self.le_angleTole.setValidator(QDoubleValidator())
        self.le_bandWidth.setValidator(QDoubleValidator())
        self.le_lagInterval.setValidator(QDoubleValidator())
        self.le_lagTole.setValidator(QDoubleValidator())
        self.le_lagNum.setValidator(QIntValidator())
        self.le_bx.setValidator(QDoubleValidator())
        self.le_by.setValidator(QDoubleValidator())
        self.le_bz.setValidator(QDoubleValidator())
        self.le_ex.setValidator(QDoubleValidator())
        self.le_ey.setValidator(QDoubleValidator())
        self.le_ez.setValidator(QDoubleValidator())
        self.le_nx.setValidator(QIntValidator())
        self.le_ny.setValidator(QIntValidator())
        self.le_nz.setValidator(QIntValidator())

        self.comb_fieldZ.setEnabled(self.check_fieldZ.isChecked())
        self.gb_variogram.setVisible(False)
        self.gb_customVariogram.setVisible(False)
        self.setLagsSearchMode(False)
        self.comb_variogramModel.insertItems(0, [key for key in VariogramModelMap])
        self.comb_method.insertItems(0, ['IDW', 'SK', 'OK'])
        cpus = tf.config.experimental.list_physical_devices('CPU')
        gpus = tf.config.experimental.list_physical_devices('GPU')
        for cpu in cpus:
            name = cpu.name[cpu.name.find(':') + 1:]
            self.comb_device.addItem(name)
        for gpu in gpus:
            name = gpu.name[gpu.name.find(':') + 1:]
            self.comb_device.addItem(name)

    def dataChanged(self):
        fields = self.data.columns.tolist()
        self.comb_fieldX.clear()
        self.comb_fieldX.insertItems(0, fields)
        self.comb_fieldX.setCurrentIndex(0)
        self.comb_fieldY.clear()
        self.comb_fieldY.insertItems(0, fields)
        self.comb_fieldY.setCurrentIndex(1)
        self.comb_fieldZ.clear()
        self.comb_fieldZ.insertItems(0, fields)
        idx = 2
        if self.check_fieldZ.isChecked():
            self.comb_fieldY.setCurrentIndex(idx)
            idx += 1
        self.comb_fieldP.clear()
        self.comb_fieldP.insertItems(0, fields)
        self.comb_fieldP.setCurrentIndex(idx)

    def setAlphaVisibility(self, visible):
        self.label_alpha.setVisible(visible)
        self.le_alpha.setVisible(visible)

    def setLagsSearchMode(self, is3d):
        self.label_angleDip.setVisible(is3d)
        self.le_angleDip.setVisible(is3d)
        if is3d:
            self.label_angle.setText("azimuth angle")
        else:
            self.label_angle.setText("angle")

    def onloadFileClicked(self):
        dir = '.'
        curFileName = self.le_loadFile.text()
        if curFileName != "":
            dir = curFileName[:curFileName.rfind('/')]
        fileName, _ = QFileDialog.getOpenFileName(self, 'load file', dir, 'file (*.csv *.gslib)')
        if fileName == '':
            return
        self.le_loadFile.setText(fileName)
        if fileName.endswith('csv'):
            self.data = pd.read_csv(fileName)
        else:
            self.data = readGslibPoints(fileName)
        self.dataChanged()

    def setupSamples(self):
        if self.data is None:
            return
        cols = [self.comb_fieldX.currentText(), self.comb_fieldY.currentText(), self.comb_fieldP.currentText()]
        if self.check_fieldZ.isChecked():
            cols.insert(2, self.comb_fieldZ.currentText())
        self.samples = self.data[cols].values
        self.cols = cols

    def onCalcVariogramClicked(self):
        self.statusBar.showMessage("Processing...",10000)#show message in  status bar, duration time set to 10s.
        begin=time.perf_counter()
        self.setupSamples()
        if self.samples is None:
            return
        if self.rb_calcVariogramCustom.isChecked():
            if self.check_fieldZ.isChecked():
                try:
                    lagNum = eval(self.le_lagNum.text())
                    lagInterval = eval(self.le_lagInterval.text())
                    lagTole = eval(self.le_lagTole.text())
                    azimuth = eval(self.le_angle.text())
                    dip = eval(self.le_angleDip.text())
                    angleTole = eval(self.le_angleTole.text())
                    bandWidth = eval(self.le_bandWidth.text())
                except:
                    QMessageBox.critical(self, "Error", "Missing parameters!")
                    return
                vecs = calcVecs(self.samples, repeat=False)
                habv = calcHABVByVecs(vecs)
                lags, _ = search3d(vecs[:, :3], habv[:, 3], lagNum, lagInterval, lagTole, azimuth, dip, angleTole,
                                   bandWidth)
                if len(lags) < 3:
                    QMessageBox.critical(self, "Warning", "Lags number less than 3! Using all pairs.")
                    lags = calcHV(self.samples, 3)
                self.vb = VariogramBuilder(lags, self.variogramModel())
            else:
                try:
                    lagNum = eval(self.le_lagNum.text())
                    lagInterval = eval(self.le_lagInterval.text())
                    lagTole = eval(self.le_lagTole.text())
                    angle = eval(self.le_angle.text())
                    angleTole = eval(self.le_angleTole.text())
                    bandWidth = eval(self.le_bandWidth.text())
                except:
                    QMessageBox.critical(self, "Error", "Missing parameters!")
                    return
                vecs = calcVecs(self.samples, repeat=False)
                hav = calcHAVByVecs(vecs)
                lags, _ = search2d(vecs[:, :2], hav[:, 2], lagNum, lagInterval, lagTole, angle, angleTole, bandWidth)
                if len(lags) < 3:
                    QMessageBox.critical(self, "Warning", "Lags number less than 3! Using all pairs.")
                    lags = calcHV(self.samples)
                self.vb = VariogramBuilder(lags, self.variogramModel())
        elif self.rb_calcVariogramDefault.isChecked():
            if self.check_fieldZ.isChecked():
                self.vb = calculateDefaultVariogram3D(self.samples, self.variogramModel())
            else:
                self.vb = calculateDefaultVariogram2D(self.samples, self.variogramModel())
        else:
            if self.check_fieldZ.isChecked():
                self.vb = calculateOmnidirectionalVariogram3D(self.samples)
            else:
                self.vb = calculateOmnidirectionalVariogram2D(self.samples)
        if type(self.vb) == tuple:
            self.canvasVariogram.fig.clear()
            nestVariogram, variogramBuilders = self.vb
            unitVectors = nestVariogram.unitVectors
            col = 3
            row = int(np.ceil(len(unitVectors) / col))
            for idx, vb in enumerate(variogramBuilders):
                if len(unitVectors[0]) == 3:
                    dip = np.arcsin(unitVectors[idx][2])
                    if np.cos(dip) == 0.0:
                        azimuth = 0.0
                    else:
                        azimuth = np.arccos(unitVectors[idx][0] / np.cos(dip))
                    text = 'azimuth: {:.4f}, dip: {:.4f}'.format(azimuth, dip)
                else:
                    azimuth = np.arccos(unitVectors[idx][0])
                    text = 'azimuth: {:.4f}'.format(azimuth)
                axes = self.canvasVariogram.fig.add_subplot(row, col, idx + 1)
                vb.showVariogram(axes)
                axes.legend()
                axes.set_title(text)
                axes.set_xlabel('distance')  # ,loc='right')
                axes.set_ylabel('variance')  # ,loc='top')
            self.canvasVariogram.fig.subplots_adjust(wspace=0.5, hspace=0.7)
            self.canvasVariogram.draw()
            self.canvasVariogram.show()
        else:
            self.canvasVariogram.fig.clear()
            self.axesVariogram = self.canvasVariogram.fig.add_subplot(111, )
            self.axesVariogram.cla()
            self.vb.showVariogram(self.axesVariogram)
            self.axesVariogram.set(title='Variogram', xlabel='distance', ylabel='variance')
            self.canvasVariogram.draw()
            self.canvasVariogram.show()
        self.tabWidget.setCurrentIndex(1)
        self.rb_variogramCalced.setEnabled(True)
        self.rb_variogramCalced.setChecked(True)
        end = time.perf_counter()
        t = end - begin
        self.statusBar.showMessage("Time consumed(/s): {:.4f}".format(t),10000)

    def onShowPointsClicked(self):
        if self.actorPS is not None:
            self.ren.RemoveActor(self.actorPS)
        if self.actorAxes is not None:
            self.ren.RemoveActor(self.actorAxes)
        self.setupSamples()
        if self.samples is None:
            return
        self.le_bx.setText(str(self.samples[:, 0].min()))
        self.le_by.setText(str(self.samples[:, 1].min()))
        self.le_ex.setText(str(self.samples[:, 0].max()))
        self.le_ey.setText(str(self.samples[:, 1].max()))
        if self.check_fieldZ.isChecked():
            self.le_bz.setText(str(self.samples[:, 2].min()))
            self.le_ez.setText(str(self.samples[:, 2].max()))

        points = vtk.vtkPoints()
        points.SetNumberOfPoints(self.samples.shape[0])
        scalar = vtk.vtkFloatArray()
        scalar.SetName("scalar")
        scalar.SetNumberOfComponents(1)
        scalar.SetNumberOfValues(self.samples.shape[0])
        if self.check_fieldZ.isChecked():
            bounds = [self.samples[:, 0].min(), self.samples[:, 0].max(), self.samples[:, 1].min(),
                      self.samples[:, 1].max(), self.samples[:, 2].min(), self.samples[:, 2].max()]
            scalarRange = [self.samples[:, 3].min(), self.samples[:, 3].max()]
            for idx, point in enumerate(self.samples):
                points.SetPoint(idx, point[:3])
                scalar.SetTuple1(idx, point[3])
        else:
            bounds = [self.samples[:, 0].min(), self.samples[:, 0].max(), self.samples[:, 1].min(),
                      self.samples[:, 1].max(), 0, 0]
            scalarRange = [self.samples[:, 2].min(), self.samples[:, 2].max()]
            for idx, point in enumerate(self.samples):
                points.SetPoint(idx, point[0], point[1], 0)
                scalar.SetTuple1(idx, point[2])
        polyData = vtk.vtkPolyData()
        polyData.SetPoints(points)
        polyData.GetPointData().AddArray(scalar)
        polyData.GetPointData().SetActiveScalars("scalar")
        vertexGlyphFilter = vtk.vtkVertexGlyphFilter()
        vertexGlyphFilter.AddInputData(polyData)
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(vertexGlyphFilter.GetOutputPort())
        mapper.SetScalarRange(*scalarRange)
        ctf = colorMap.getCTF(scalarRange[0], scalarRange[1], ColorMap)
        ctf.SetDiscretize(True)
        ctf.Build()
        mapper.SetLookupTable(ctf)
        self.actorScalarBar.SetLookupTable(ctf)
        self.ren.AddActor2D(self.actorScalarBar)
        self.ctf = ctf
        self.scalarRange = scalarRange

        self.actorPS = vtk.vtkActor()
        self.actorPS.SetMapper(mapper)
        self.actorPS.GetProperty().SetPointSize(5)
        self.ren.AddActor(self.actorPS)

        self.actorAxes = vtk.vtkCubeAxesActor()
        self.actorAxes.SetBounds(bounds)
        self.actorAxes.SetCamera(self.ren.GetActiveCamera())
        self.axesMarker = vtk.vtkOrientationMarkerWidget()
        self.axesMarker.SetOutlineColor(0, 0, 0)
        self.axesMarker.SetOrientationMarker(self.actorAxes)
        self.axesMarker.SetCurrentRenderer(self.ren)
        self.axesMarker.SetInteractor(self.iren)
        self.axesMarker.SetEnabled(1)
        self.axesMarker.InteractiveOff()
        self.ren.AddActor(self.actorAxes)

        self.ren.ResetCamera()
        self.vtkWidget.GetRenderWindow().Render()
        # self.axesPoints.cla()
        # mappable=self.axesPoints.scatter(self.samples[:, 0], self.samples[:, 1],c=self.samples[:,2],cmap='rainbow')
        # self.canvasPoints.figure.colorbar(mappable,ax=self.axesPoints)
        # self.canvasPoints.draw()
        # self.canvasPoints.show()
        self.tabWidget.setCurrentIndex(0)

    def onZCheckStateChanged(self, state):
        self.le_bz.setEnabled(state)
        self.le_ez.setEnabled(state)
        self.le_nz.setEnabled(state)
        if state:
            pass
        else:
            pass

    def cv(self):
        mode = '3d' if self.check_fieldZ.isChecked() else '2d'
        N = self.sb_neighSize.value()
        method = self.comb_method.currentText()
        if method == 'IDW':
            alpha = eval(self.le_alpha.text())
            if self.gb_useTF.isChecked():
                device = self.comb_device.currentText()
                exe = TFIDW(self.samples, mode)
                with tf.device('/' + device):
                    _, _, error = exe.crossValidate(N, alpha)
            else:
                exe = IDW(self.samples, mode)
                _, _, error = exe.crossValidate(N, alpha)
        else:
            if self.rb_variogramNone.isChecked():
                vb = None
            elif self.rb_variogramDefault.isChecked():
                vb = calculateDefaultVariogram2D(self.samples, self.variogramModel())
            else:
                vb = self.vb
            if self.gb_useTF.isChecked():
                Method = TFSK if method == 'SK' else TFOK
            else:
                Method = SK if method == 'SK' else OK
            exe = Method(self.samples, mode)
            if self.gb_useTF.isChecked():
                device = self.comb_device.currentText()
                with tf.device('/' + device):
                    if vb is None:
                        variogram = None
                    elif type(vb) == tuple:
                        variogram = getNestVariogramLayer(vb[1], vb[0].unitVectors)
                    else:
                        variogram = getVariogramLayer(vb)
                    _, _, error = exe.crossValidate(N, variogram)
            else:
                if vb is None:
                    variogram = None
                elif type(vb) == tuple:
                    variogram = vb[0]
                else:
                    variogram = vb.getVariogram()
                _, _, error = exe.crossValidate(N, variogram)
        idx = 2 if mode == '2d' else 3
        estimate = error + self.samples[:, idx]

        def resident(params, x, y):
            return (y - (params[0] * x + params[1])) ** 2

        res = least_squares(resident, [1, 0], args=(self.samples[:, idx], estimate))
        a, b = res.x

        self.axesCV.cla()
        self.axesCV.scatter(self.samples[:, idx], estimate, label="pairs")
        vran = [self.samples[:, idx].min(), self.samples[:, idx].max()]
        X = np.arange(vran[0], vran[1], (vran[1] - vran[0]) / 50.0)
        self.axesCV.plot(X, X, label="1:1 line", color="orange", linestyle='-')
        self.axesCV.plot(X, X * a + b, label="best linear line fit", color="red", linestyle='--')
        self.axesCV.set(title="Cross Validation", xlabel="sample values", ylabel="estimate values")
        self.axesCV.legend()
        self.canvasCV.draw()
        self.canvasCV.show()
        self.tabWidget.setCurrentIndex(2)

    def interp(self):
        mode = '3d' if self.check_fieldZ.isChecked() else '2d'
        try:
            bx = eval(self.le_bx.text())
            by = eval(self.le_by.text())
            ex = eval(self.le_ex.text())
            ey = eval(self.le_ey.text())
            nx = eval(self.le_nx.text())
            ny = eval(self.le_ny.text())
            if mode == '3d':
                bz = eval(self.le_bz.text())
                ez = eval(self.le_ez.text())
                nz = eval(self.le_nz.text())
                grid = Grid3D()
                grid.rectlinear([nx, ny, nz], [bx, ex], [by, ey], [bz, ez])
            else:
                grid = Grid2D()
                grid.rectlinear([nx, ny], [bx, ex], [by, ey])
        except:
            QMessageBox.critical(self, "Error", "Missing parameters!")
            return False
        N = self.sb_neighSize.value()
        method = self.comb_method.currentText()
        if method == 'IDW':
            alpha = eval(self.le_alpha.text())
            if self.gb_useTF.isChecked():
                device = self.comb_device.currentText()
                exe = TFIDW(self.samples, mode)
                with tf.device('/' + device):
                    grid.properties = exe.execute(grid.points(), N, alpha)
            else:
                exe = IDW(self.samples, mode)
                grid.properties = exe.execute(grid.points(), N, alpha)
            grid.sigmas = None
        else:
            if self.rb_variogramNone.isChecked():
                vb = None
            elif self.rb_variogramDefault.isChecked():
                vb = calculateDefaultVariogram2D(self.samples, self.variogramModel())
            else:
                vb = self.vb
            if self.gb_useTF.isChecked():
                Method = TFSK if method == 'SK' else TFOK
            else:
                Method = SK if method == 'SK' else OK
            exe = Method(self.samples, mode)
            if self.gb_useTF.isChecked():
                device = self.comb_device.currentText()
                with tf.device('/' + device):
                    if vb is None:
                        variogram = None
                    elif type(vb) == tuple:
                        variogram = getNestVariogramLayer(vb[1], vb[0].unitVectors)
                    else:
                        variogram = getVariogramLayer(vb)
                    grid.properties, grid.sigmas = exe.execute(grid.points(), N, variogram)
            else:
                if vb is None:
                    variogram = None
                elif type(vb) == tuple:
                    variogram = vb[0]
                else:
                    variogram = vb.getVariogram()
                grid.properties, grid.sigmas = exe.execute(grid.points(), N, variogram)
        self.grid = grid
        return True

    def showGrid(self):
        if self.actorGrid is not None:
            self.ren.RemoveActor(self.actorGrid)
        if self.actorAxes is not None:
            self.ren.RemoveActor(self.actorAxes)
        mode = '2d' if self.grid.__class__ == Grid2D else '3d'
        rgrid = vtk.vtkRectilinearGrid()
        dim = self.grid.dim if mode == '3d' else [*self.grid.dim, 1]
        rgrid.SetDimensions(dim)
        xcoords = vtk.vtkFloatArray()
        xcoords.SetNumberOfValues(dim[0])
        xcoords.SetNumberOfComponents(1)
        ycoords = vtk.vtkFloatArray()
        ycoords.SetNumberOfValues(dim[1])
        ycoords.SetNumberOfComponents(1)
        zcoords = vtk.vtkFloatArray()
        zcoords.SetNumberOfValues(dim[2])
        zcoords.SetNumberOfComponents(1)
        for idx, value in enumerate(self.grid.x):
            xcoords.SetValue(idx, value)
        for idx, value in enumerate(self.grid.y):
            ycoords.SetValue(idx, value)
        z = [0] if mode == '2d' else self.grid.z
        for idx, value in enumerate(z):
            zcoords.SetValue(idx, value)
        rgrid.SetXCoordinates(xcoords)
        rgrid.SetYCoordinates(ycoords)
        rgrid.SetZCoordinates(zcoords)

        scalar = vtk.vtkFloatArray()
        scalar.SetName('properties')
        scalar.SetNumberOfComponents(1)
        scalar.SetNumberOfValues(len(self.grid.properties))
        for idx, value in enumerate(self.grid.properties):
            scalar.SetValue(idx, value)
        rgrid.GetPointData().AddArray(scalar)
        if self.grid.sigmas is not None:
            scalar = vtk.vtkFloatArray()
            scalar.SetName('sigmas')
            scalar.SetNumberOfComponents(1)
            scalar.SetNumberOfValues(len(self.grid.sigmas))
            for idx, value in enumerate(self.grid.sigmas):
                scalar.SetValue(idx, value)
            rgrid.GetPointData().AddArray(scalar)

        rgrid.GetPointData().SetActiveScalars("properties")

        mapper = vtk.vtkDataSetMapper()
        mapper.SetInputData(rgrid)
        mapper.SetScalarRange(*self.scalarRange)
        mapper.SetLookupTable(self.ctf)

        self.actorGrid = vtk.vtkActor()
        self.actorGrid.SetMapper(mapper)
        self.actorGrid.GetProperty().SetInterpolationToGouraud()
        self.ren.AddActor(self.actorGrid)

        bounds = [*self.grid.xran, *self.grid.yran]
        if mode == '2d':
            bounds.extend([0, 0])
        else:
            bounds.extend([*self.grid.zran])
        self.actorAxes = vtk.vtkCubeAxesActor()
        self.actorAxes.SetBounds(bounds)
        self.actorAxes.SetCamera(self.ren.GetActiveCamera())
        self.axesMarker = vtk.vtkOrientationMarkerWidget()
        self.axesMarker.SetOutlineColor(0, 0, 0)
        self.axesMarker.SetOrientationMarker(self.actorAxes)
        self.axesMarker.SetCurrentRenderer(self.ren)
        self.axesMarker.SetInteractor(self.iren)
        self.axesMarker.SetEnabled(1)
        self.axesMarker.InteractiveOff()
        self.ren.AddActor(self.actorAxes)

        self.ren.ResetCamera()
        self.vtkWidget.GetRenderWindow().Render()
        self.tabWidget.setCurrentIndex(0)

    def onExecuteInterpClicked(self):
        self.setupSamples()
        if self.samples is None:
            QMessageBox.critical(self, "Error", "No sample data!")
            return
        self.statusBar.showMessage("Processing...", 5000)
        begin = time.perf_counter()
        if self.rb_interp.isChecked():
            if not self.interp():
                return
            self.showGrid()
        else:
            self.cv()
        end = time.perf_counter()
        t = end - begin
        self.statusBar.showMessage("Time consumed(/s): {:.4f}".format(t),5000)

    def variogramModel(self):
        return self.comb_variogramModel.currentText()

    def connectSignals(self):
        self.pb_loadFile.clicked.connect(self.onloadFileClicked)
        self.check_fieldZ.stateChanged.connect(lambda state: self.comb_fieldZ.setEnabled(state))
        self.check_fieldZ.stateChanged.connect(self.onZCheckStateChanged)
        self.check_fieldZ.stateChanged.connect(self.setLagsSearchMode)
        self.pb_calcVariogram.clicked.connect(self.onCalcVariogramClicked)
        self.pb_showPoints.clicked.connect(self.onShowPointsClicked)
        self.rb_interp.toggled.connect(lambda state: self.gb_grid.setVisible(state))
        self.pb_executeInterp.clicked.connect(self.onExecuteInterpClicked)
        self.comb_method.currentTextChanged.connect(lambda text: self.setAlphaVisibility(text == 'IDW'))
        self.comb_method.currentTextChanged.connect(lambda text: self.gb_variogram.setVisible(text != 'IDW'))
        self.rb_calcVariogramCustom.toggled.connect(lambda state: self.gb_customVariogram.setVisible(state))
        self.pb_savePs.clicked.connect(self.onSavePsClicked)
        self.pb_saveGrid.clicked.connect(self.onSaveGridClicked)

    def onSavePsClicked(self):
        self.setupSamples()
        if self.samples is None:
            QMessageBox.critical(self, "Error", "No sample data!")
            return
        path, _ = QFileDialog.getSaveFileName(self, "Save points, gslib or vtk", '', 'vtk(*.vtk);;gslib(*.gslib)')
        if path == '':
            return
        if path.endswith('vtk'):
            saveVTKPoints(path, self.actorPS.GetMapper().GetInput())
        else:
            saveGslibPoints(path, self.cols, self.samples)

    def onSaveGridClicked(self):
        if self.grid is None:
            QMessageBox.critical(self, "Error", "No grid!")
            return
        path, _ = QFileDialog.getSaveFileName(self, "Save grid, gslib or vtk", '', 'vtk(*.vtk);;gslib(*.gslib)')
        if path == '':
            return
        if path.endswith('vtk'):
            saveVTKGrid(path, self.actorGrid.GetMapper().GetInput())
        else:
            begin = [self.grid.xran[0], self.grid.yran[0]]
            step = [(self.grid.xran[1] - self.grid.xran[0]) / (self.grid.dim[0] - 1),
                    (self.grid.yran[1] - self.grid.yran[0]) / (self.grid.dim[1] - 1)]
            if self.grid.__class__ == Grid3D:
                begin.append(self.grid.zran[0])
                step.append((self.grid.zran[1] - self.grid.zran[0]) / (self.grid.dim[2] - 1))
            colNames = ['properties']
            if self.grid.sigmas is not None:
                colNames.append('sigmas')
                data = np.concatenate([self.grid.properties, self.grid.sigmas], axis=1)
            else:
                data = self.grid.properties
            saveGslibGrid(path, self.grid.dim, begin, step, colNames, data)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    exe = Tool()
    exe.showMaximized()
    sys.exit(app.exec_())



from tfinterpy.krige import SK
from tfinterpy.grid import Grid2D
from tfinterpy.variogram import calculateDefaultVariogram2D
from examples.ncFileUtil import getSamples
from examples.plotUtils import *
import numpy as np
import random

if __name__ == "__main__":
    random.seed(1)
    filePath = "data/tatitlek_815_mhhw_2011.nc"
    W, H = 100, 100
    M = 100
    offset = (400, 800)
    N = 8

    # Get sample points and original elevation data from netcdf file
    # containing Digital Elevation Model data.
    samples, lats, lons, ele = getSamples(filePath, 'lat', 'lon', 'Band1', offset, (W, H), M)

    # Create linear 2D grid.
    grid = Grid2D()
    grid.rectlinear((W, H), (1, W), (1, H))

    # Calculate a default variogram function.
    vb = calculateDefaultVariogram2D(samples)
    exe = SK(samples)# Create a sk interpolator.
    # Perform interpolation of all points in the grid.
    grid.pro, grid.sigma = exe.execute(grid.points(), N, vb.getVariogram())

    print(exe.crossValidate(N, vb.getVariogram()))# Perform leave-one-out validation and print result.
    print(exe.crossValidateKFold(10, N, vb.getVariogram()))# Perform k-fold validation and print result.

    grid.pro.resize((grid.dim[1], grid.dim[0]))# Reshape properties as 2d ndarray.
    grid.sigma.resize((grid.dim[1], grid.dim[0]))# Reshape kriging variances as 2d ndarray.
    ae = np.abs(grid.pro - ele)
    print("mae:", np.mean(ae))
    plt.subplot(221)
    img2d(ele, title="origin")# Plotting original elevation data.
    plt.subplot(222)
    img2d(grid.pro, title="estimate")# Plotting interpolation results.
    plt.subplot(223)
    img2d(ae, title="error")# Plotting absolute errors.
    plt.subplot(224)
    img2d(grid.sigma, title="krige variance")# Plotting kriging variances.
    plt.show()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/tool.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1274, 1096)
        font = QtGui.QFont()
        font.setFamily("Times New Roman")
        font.setPointSize(12)
        MainWindow.setFont(font)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout_18 = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout_18.setObjectName("horizontalLayout_18")
        self.scrollArea = QtWidgets.QScrollArea(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy)
        self.scrollArea.setMaximumSize(QtCore.QSize(385, 16777215))
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 362, 1476))
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollAreaWidgetContents.sizePolicy().hasHeightForWidth())
        self.scrollAreaWidgetContents.setSizePolicy(sizePolicy)
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.horizontalLayout_17 = QtWidgets.QHBoxLayout(self.scrollAreaWidgetContents)
        self.horizontalLayout_17.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_17.setSpacing(0)
        self.horizontalLayout_17.setObjectName("horizontalLayout_17")
        self.widget = QtWidgets.QWidget(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setObjectName("widget")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.widget)
        self.verticalLayout_4.setContentsMargins(5, 5, 5, 5)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_6 = QtWidgets.QLabel(self.widget)
        font = QtGui.QFont()
        font.setFamily("Times New Roman")
        font.setPointSize(11)
        self.label_6.setFont(font)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_4.addWidget(self.label_6)
        self.line_3 = QtWidgets.QFrame(self.widget)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.line_3.setLineWidth(1)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setObjectName("line_3")
        self.verticalLayout_4.addWidget(self.line_3)
        self.widget_4 = QtWidgets.QWidget(self.widget)
        self.widget_4.setObjectName("widget_4")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.widget_4)
        self.verticalLayout_9.setContentsMargins(-1, 0, 0, 0)
        self.verticalLayout_9.setSpacing(4)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pb_loadFile = QtWidgets.QPushButton(self.widget_4)
        self.pb_loadFile.setObjectName("pb_loadFile")
        self.horizontalLayout.addWidget(self.pb_loadFile)
        self.le_loadFile = QtWidgets.QLineEdit(self.widget_4)
        self.le_loadFile.setReadOnly(True)
        self.le_loadFile.setObjectName("le_loadFile")
        self.horizontalLayout.addWidget(self.le_loadFile)
        self.verticalLayout_9.addLayout(self.horizontalLayout)
        self.groupBox = QtWidgets.QGroupBox(self.widget_4)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.comb_fieldX = QtWidgets.QComboBox(self.groupBox)
        self.comb_fieldX.setObjectName("comb_fieldX")
        self.horizontalLayout_2.addWidget(self.comb_fieldX)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_2 = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_3.addWidget(self.label_2)
        self.comb_fieldY = QtWidgets.QComboBox(self.groupBox)
        self.comb_fieldY.setObjectName("comb_fieldY")
        self.horizontalLayout_3.addWidget(self.comb_fieldY)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_3 = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_4.addWidget(self.label_3)
        self.comb_fieldZ = QtWidgets.QComboBox(self.groupBox)
        self.comb_fieldZ.setObjectName("comb_fieldZ")
        self.horizontalLayout_4.addWidget(self.comb_fieldZ)
        self.check_fieldZ = QtWidgets.QCheckBox(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.check_fieldZ.sizePolicy().hasHeightForWidth())
        self.check_fieldZ.setSizePolicy(sizePolicy)
        self.check_fieldZ.setText("")
        self.check_fieldZ.setObjectName("check_fieldZ")
        self.horizontalLayout_4.addWidget(self.check_fieldZ)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_4 = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_5.addWidget(self.label_4)
        self.comb_fieldP = QtWidgets.QComboBox(self.groupBox)
        self.comb_fieldP.setObjectName("comb_fieldP")
        self.horizontalLayout_5.addWidget(self.comb_fieldP)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.pb_showPoints = QtWidgets.QPushButton(self.groupBox)
        self.pb_showPoints.setObjectName("pb_showPoints")
        self.verticalLayout.addWidget(self.pb_showPoints)
        self.verticalLayout_9.addWidget(self.groupBox)
        self.verticalLayout_4.addWidget(self.widget_4)
        self.label_7 = QtWidgets.QLabel(self.widget)
        font = QtGui.QFont()
        font.setFamily("Times New Roman")
        font.setPointSize(10)
        self.label_7.setFont(font)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_4.addWidget(self.label_7)
        self.line = QtWidgets.QFrame(self.widget)
        self.line.setFrameShadow(QtWidgets.QFrame.Raised)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout_4.addWidget(self.line)
        self.widget_3 = QtWidgets.QWidget(self.widget)
        self.widget_3.setObjectName("widget_3")
        self.horizontalLayout_16 = QtWidgets.QHBoxLayout(self.widget_3)
        self.horizontalLayout_16.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_16.setObjectName("horizontalLayout_16")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.pb_calcVariogram = QtWidgets.QPushButton(self.widget_3)
        self.pb_calcVariogram.setObjectName("pb_calcVariogram")
        self.verticalLayout_3.addWidget(self.pb_calcVariogram)
        self.groupBox_2 = QtWidgets.QGroupBox(self.widget_3)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_11 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_11.setObjectName("verticalLayout_11")
        self.rb_calcVariogramDefault = QtWidgets.QRadioButton(self.groupBox_2)
        self.rb_calcVariogramDefault.setChecked(True)
        self.rb_calcVariogramDefault.setObjectName("rb_calcVariogramDefault")
        self.verticalLayout_11.addWidget(self.rb_calcVariogramDefault)
        self.rb_calcVariogramCustom = QtWidgets.QRadioButton(self.groupBox_2)
        self.rb_calcVariogramCustom.setEnabled(True)
        self.rb_calcVariogramCustom.setObjectName("rb_calcVariogramCustom")
        self.verticalLayout_11.addWidget(self.rb_calcVariogramCustom)
        self.rb_calcVariogramOmni = QtWidgets.QRadioButton(self.groupBox_2)
        self.rb_calcVariogramOmni.setEnabled(True)
        self.rb_calcVariogramOmni.setObjectName("rb_calcVariogramOmni")
        self.verticalLayout_11.addWidget(self.rb_calcVariogramOmni)
        self.verticalLayout_3.addWidget(self.groupBox_2)
        self.horizontalLayout_14 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_14.setObjectName("horizontalLayout_14")
        self.label_5 = QtWidgets.QLabel(self.widget_3)
        self.label_5.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_14.addWidget(self.label_5)
        self.comb_variogramModel = QtWidgets.QComboBox(self.widget_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comb_variogramModel.sizePolicy().hasHeightForWidth())
        self.comb_variogramModel.setSizePolicy(sizePolicy)
        self.comb_variogramModel.setObjectName("comb_variogramModel")
        self.horizontalLayout_14.addWidget(self.comb_variogramModel)
        self.horizontalLayout_14.setStretch(0, 1)
        self.horizontalLayout_14.setStretch(1, 2)
        self.verticalLayout_3.addLayout(self.horizontalLayout_14)
        self.gb_customVariogram = QtWidgets.QGroupBox(self.widget_3)
        self.gb_customVariogram.setCheckable(False)
        self.gb_customVariogram.setChecked(False)
        self.gb_customVariogram.setObjectName("gb_customVariogram")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.gb_customVariogram)
        self.verticalLayout_2.setContentsMargins(4, 4, 4, 4)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_lagInterval = QtWidgets.QLabel(self.gb_customVariogram)
        self.label_lagInterval.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_lagInterval.setObjectName("label_lagInterval")
        self.verticalLayout_6.addWidget(self.label_lagInterval)
        self.label_lagNum = QtWidgets.QLabel(self.gb_customVariogram)
        self.label_lagNum.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_lagNum.setObjectName("label_lagNum")
        self.verticalLayout_6.addWidget(self.label_lagNum)
        self.label_lagTole = QtWidgets.QLabel(self.gb_customVariogram)
        self.label_lagTole.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_lagTole.setObjectName("label_lagTole")
        self.verticalLayout_6.addWidget(self.label_lagTole)
        self.label_angle = QtWidgets.QLabel(self.gb_customVariogram)
        self.label_angle.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_angle.setObjectName("label_angle")
        self.verticalLayout_6.addWidget(self.label_angle)
        self.label_angleDip = QtWidgets.QLabel(self.gb_customVariogram)
        self.label_angleDip.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_angleDip.setObjectName("label_angleDip")
        self.verticalLayout_6.addWidget(self.label_angleDip)
        self.label_angleTole = QtWidgets.QLabel(self.gb_customVariogram)
        self.label_angleTole.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_angleTole.setObjectName("label_angleTole")
        self.verticalLayout_6.addWidget(self.label_angleTole)
        self.label_bandWidth = QtWidgets.QLabel(self.gb_customVariogram)
        self.label_bandWidth.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_bandWidth.setObjectName("label_bandWidth")
        self.verticalLayout_6.addWidget(self.label_bandWidth)
        self.horizontalLayout_6.addLayout(self.verticalLayout_6)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.le_lagInterval = QtWidgets.QLineEdit(self.gb_customVariogram)
        self.le_lagInterval.setObjectName("le_lagInterval")
        self.verticalLayout_5.addWidget(self.le_lagInterval)
        self.le_lagNum = QtWidgets.QLineEdit(self.gb_customVariogram)
        self.le_lagNum.setObjectName("le_lagNum")
        self.verticalLayout_5.addWidget(self.le_lagNum)
        self.le_lagTole = QtWidgets.QLineEdit(self.gb_customVariogram)
        self.le_lagTole.setObjectName("le_lagTole")
        self.verticalLayout_5.addWidget(self.le_lagTole)
        self.le_angle = QtWidgets.QLineEdit(self.gb_customVariogram)
        self.le_angle.setObjectName("le_angle")
        self.verticalLayout_5.addWidget(self.le_angle)
        self.le_angleDip = QtWidgets.QLineEdit(self.gb_customVariogram)
        self.le_angleDip.setObjectName("le_angleDip")
        self.verticalLayout_5.addWidget(self.le_angleDip)
        self.le_angleTole = QtWidgets.QLineEdit(self.gb_customVariogram)
        self.le_angleTole.setObjectName("le_angleTole")
        self.verticalLayout_5.addWidget(self.le_angleTole)
        self.le_bandWidth = QtWidgets.QLineEdit(self.gb_customVariogram)
        self.le_bandWidth.setObjectName("le_bandWidth")
        self.verticalLayout_5.addWidget(self.le_bandWidth)
        self.horizontalLayout_6.addLayout(self.verticalLayout_5)
        self.verticalLayout_2.addLayout(self.horizontalLayout_6)
        self.verticalLayout_3.addWidget(self.gb_customVariogram)
        self.horizontalLayout_16.addLayout(self.verticalLayout_3)
        self.verticalLayout_4.addWidget(self.widget_3)
        self.label_8 = QtWidgets.QLabel(self.widget)
        font = QtGui.QFont()
        font.setFamily("Times New Roman")
        font.setPointSize(10)
        self.label_8.setFont(font)
        self.label_8.setObjectName("label_8")
        self.verticalLayout_4.addWidget(self.label_8)
        self.line_2 = QtWidgets.QFrame(self.widget)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout_4.addWidget(self.line_2)
        self.widget_2 = QtWidgets.QWidget(self.widget)
        self.widget_2.setObjectName("widget_2")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.widget_2)
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.pb_executeInterp = QtWidgets.QPushButton(self.widget_2)
        self.pb_executeInterp.setObjectName("pb_executeInterp")
        self.verticalLayout_7.addWidget(self.pb_executeInterp)
        self.gb_mode = QtWidgets.QGroupBox(self.widget_2)
        self.gb_mode.setFlat(True)
        self.gb_mode.setCheckable(False)
        self.gb_mode.setChecked(False)
        self.gb_mode.setObjectName("gb_mode")
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout(self.gb_mode)
        self.horizontalLayout_10.setContentsMargins(4, 4, 4, 4)
        self.horizontalLayout_10.setSpacing(3)
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout()
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.rb_interp = QtWidgets.QRadioButton(self.gb_mode)
        self.rb_interp.setChecked(True)
        self.rb_interp.setObjectName("rb_interp")
        self.verticalLayout_8.addWidget(self.rb_interp)
        self.rb_cv = QtWidgets.QRadioButton(self.gb_mode)
        self.rb_cv.setObjectName("rb_cv")
        self.verticalLayout_8.addWidget(self.rb_cv)
        self.horizontalLayout_10.addLayout(self.verticalLayout_8)
        self.verticalLayout_7.addWidget(self.gb_mode)
        self.gb_grid = QtWidgets.QGroupBox(self.widget_2)
        self.gb_grid.setFlat(True)
        self.gb_grid.setObjectName("gb_grid")
        self.verticalLayout_10 = QtWidgets.QVBoxLayout(self.gb_grid)
        self.verticalLayout_10.setContentsMargins(4, 4, 4, 4)
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.le_ez = QtWidgets.QLineEdit(self.gb_grid)
        self.le_ez.setEnabled(False)
        self.le_ez.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_ez.setObjectName("le_ez")
        self.gridLayout.addWidget(self.le_ez, 2, 3, 1, 1)
        self.le_ey = QtWidgets.QLineEdit(self.gb_grid)
        self.le_ey.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_ey.setObjectName("le_ey")
        self.gridLayout.addWidget(self.le_ey, 2, 2, 1, 1)
        self.label_10 = QtWidgets.QLabel(self.gb_grid)
        self.label_10.setAlignment(QtCore.Qt.AlignCenter)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 0, 2, 1, 1)
        self.le_by = QtWidgets.QLineEdit(self.gb_grid)
        self.le_by.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_by.setObjectName("le_by")
        self.gridLayout.addWidget(self.le_by, 1, 2, 1, 1)
        self.label_14 = QtWidgets.QLabel(self.gb_grid)
        self.label_14.setAlignment(QtCore.Qt.AlignCenter)
        self.label_14.setObjectName("label_14")
        self.gridLayout.addWidget(self.label_14, 0, 3, 1, 1)
        self.le_bx = QtWidgets.QLineEdit(self.gb_grid)
        self.le_bx.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_bx.setObjectName("le_bx")
        self.gridLayout.addWidget(self.le_bx, 1, 1, 1, 1)
        self.le_bz = QtWidgets.QLineEdit(self.gb_grid)
        self.le_bz.setEnabled(False)
        self.le_bz.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_bz.setObjectName("le_bz")
        self.gridLayout.addWidget(self.le_bz, 1, 3, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.gb_grid)
        self.label_9.setAlignment(QtCore.Qt.AlignCenter)
        self.label_9.setObjectName("label_9")
        self.gridLayout.addWidget(self.label_9, 0, 1, 1, 1)
        self.label_16 = QtWidgets.QLabel(self.gb_grid)
        self.label_16.setAlignment(QtCore.Qt.AlignCenter)
        self.label_16.setObjectName("label_16")
        self.gridLayout.addWidget(self.label_16, 1, 0, 1, 1)
        self.le_ex = QtWidgets.QLineEdit(self.gb_grid)
        self.le_ex.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_ex.setObjectName("le_ex")
        self.gridLayout.addWidget(self.le_ex, 2, 1, 1, 1)
        self.le_nx = QtWidgets.QLineEdit(self.gb_grid)
        self.le_nx.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_nx.setObjectName("le_nx")
        self.gridLayout.addWidget(self.le_nx, 3, 1, 1, 1)
        self.le_ny = QtWidgets.QLineEdit(self.gb_grid)
        self.le_ny.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_ny.setObjectName("le_ny")
        self.gridLayout.addWidget(self.le_ny, 3, 2, 1, 1)
        self.le_nz = QtWidgets.QLineEdit(self.gb_grid)
        self.le_nz.setEnabled(False)
        self.le_nz.setMaximumSize(QtCore.QSize(60, 16777215))
        self.le_nz.setObjectName("le_nz")
        self.gridLayout.addWidget(self.le_nz, 3, 3, 1, 1)
        self.label_17 = QtWidgets.QLabel(self.gb_grid)
        self.label_17.setAlignment(QtCore.Qt.AlignCenter)
        self.label_17.setObjectName("label_17")
        self.gridLayout.addWidget(self.label_17, 2, 0, 1, 1)
        self.label_18 = QtWidgets.QLabel(self.gb_grid)
        self.label_18.setAlignment(QtCore.Qt.AlignCenter)
        self.label_18.setObjectName("label_18")
        self.gridLayout.addWidget(self.label_18, 3, 0, 1, 1)
        self.verticalLayout_10.addLayout(self.gridLayout)
        self.verticalLayout_7.addWidget(self.gb_grid)
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.label_11 = QtWidgets.QLabel(self.widget_2)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout_12.addWidget(self.label_11)
        self.comb_method = QtWidgets.QComboBox(self.widget_2)
        self.comb_method.setObjectName("comb_method")
        self.horizontalLayout_12.addWidget(self.comb_method)
        self.verticalLayout_7.addLayout(self.horizontalLayout_12)
        self.gb_variogram = QtWidgets.QGroupBox(self.widget_2)
        self.gb_variogram.setObjectName("gb_variogram")
        self.horizontalLayout_15 = QtWidgets.QHBoxLayout(self.gb_variogram)
        self.horizontalLayout_15.setContentsMargins(4, 4, 4, 4)
        self.horizontalLayout_15.setSpacing(4)
        self.horizontalLayout_15.setObjectName("horizontalLayout_15")
        self.rb_variogramNone = QtWidgets.QRadioButton(self.gb_variogram)
        self.rb_variogramNone.setObjectName("rb_variogramNone")
        self.horizontalLayout_15.addWidget(self.rb_variogramNone)
        self.rb_variogramDefault = QtWidgets.QRadioButton(self.gb_variogram)
        self.rb_variogramDefault.setChecked(True)
        self.rb_variogramDefault.setObjectName("rb_variogramDefault")
        self.horizontalLayout_15.addWidget(self.rb_variogramDefault)
        self.rb_variogramCalced = QtWidgets.QRadioButton(self.gb_variogram)
        self.rb_variogramCalced.setEnabled(False)
        self.rb_variogramCalced.setObjectName("rb_variogramCalced")
        self.horizontalLayout_15.addWidget(self.rb_variogramCalced)
        self.verticalLayout_7.addWidget(self.gb_variogram)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.label_12 = QtWidgets.QLabel(self.widget_2)
        self.label_12.setObjectName("label_12")
        self.horizontalLayout_7.addWidget(self.label_12)
        self.sb_neighSize = QtWidgets.QSpinBox(self.widget_2)
        self.sb_neighSize.setMinimum(2)
        self.sb_neighSize.setMaximum(99)
        self.sb_neighSize.setProperty("value", 8)
        self.sb_neighSize.setObjectName("sb_neighSize")
        self.horizontalLayout_7.addWidget(self.sb_neighSize)
        self.verticalLayout_7.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.label_alpha = QtWidgets.QLabel(self.widget_2)
        self.label_alpha.setObjectName("label_alpha")
        self.horizontalLayout_8.addWidget(self.label_alpha)
        self.le_alpha = QtWidgets.QLineEdit(self.widget_2)
        self.le_alpha.setObjectName("le_alpha")
        self.horizontalLayout_8.addWidget(self.le_alpha)
        self.verticalLayout_7.addLayout(self.horizontalLayout_8)
        self.gb_useTF = QtWidgets.QGroupBox(self.widget_2)
        self.gb_useTF.setFlat(True)
        self.gb_useTF.setCheckable(True)
        self.gb_useTF.setObjectName("gb_useTF")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout(self.gb_useTF)
        self.horizontalLayout_9.setContentsMargins(4, 4, 4, 4)
        self.horizontalLayout_9.setSpacing(4)
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        self.label_13 = QtWidgets.QLabel(self.gb_useTF)
        self.label_13.setObjectName("label_13")
        self.horizontalLayout_13.addWidget(self.label_13)
        self.comb_device = QtWidgets.QComboBox(self.gb_useTF)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comb_device.sizePolicy().hasHeightForWidth())
        self.comb_device.setSizePolicy(sizePolicy)
        self.comb_device.setObjectName("comb_device")
        self.horizontalLayout_13.addWidget(self.comb_device)
        self.horizontalLayout_9.addLayout(self.horizontalLayout_13)
        self.verticalLayout_7.addWidget(self.gb_useTF)
        self.verticalLayout_4.addWidget(self.widget_2)
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.pb_savePs = QtWidgets.QPushButton(self.widget)
        self.pb_savePs.setObjectName("pb_savePs")
        self.horizontalLayout_11.addWidget(self.pb_savePs)
        self.pb_saveGrid = QtWidgets.QPushButton(self.widget)
        self.pb_saveGrid.setObjectName("pb_saveGrid")
        self.horizontalLayout_11.addWidget(self.pb_saveGrid)
        self.verticalLayout_4.addLayout(self.horizontalLayout_11)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem)
        self.horizontalLayout_17.addWidget(self.widget)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.horizontalLayout_18.addWidget(self.scrollArea)
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setObjectName("tabWidget")
        self.tab_interp = QtWidgets.QWidget()
        self.tab_interp.setObjectName("tab_interp")
        self.tabWidget.addTab(self.tab_interp, "")
        self.tab_variogram = QtWidgets.QWidget()
        self.tab_variogram.setObjectName("tab_variogram")
        self.verticalLayout_12 = QtWidgets.QVBoxLayout(self.tab_variogram)
        self.verticalLayout_12.setContentsMargins(5, 5, 5, 5)
        self.verticalLayout_12.setObjectName("verticalLayout_12")
        self.scrollArea_2 = QtWidgets.QScrollArea(self.tab_variogram)
        self.scrollArea_2.setWidgetResizable(True)
        self.scrollArea_2.setObjectName("scrollArea_2")
        self.scrollAreaWidgetContents_2 = QtWidgets.QWidget()
        self.scrollAreaWidgetContents_2.setGeometry(QtCore.QRect(0, 0, 842, 1000))
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollAreaWidgetContents_2.sizePolicy().hasHeightForWidth())
        self.scrollAreaWidgetContents_2.setSizePolicy(sizePolicy)
        self.scrollAreaWidgetContents_2.setObjectName("scrollAreaWidgetContents_2")
        self.horizontalLayout_19 = QtWidgets.QHBoxLayout(self.scrollAreaWidgetContents_2)
        self.horizontalLayout_19.setContentsMargins(4, 4, 4, 4)
        self.horizontalLayout_19.setObjectName("horizontalLayout_19")
        self.vl_variogram = QtWidgets.QVBoxLayout()
        self.vl_variogram.setObjectName("vl_variogram")
        self.horizontalLayout_19.addLayout(self.vl_variogram)
        self.scrollArea_2.setWidget(self.scrollAreaWidgetContents_2)
        self.verticalLayout_12.addWidget(self.scrollArea_2)
        self.tabWidget.addTab(self.tab_variogram, "")
        self.tab_cv = QtWidgets.QWidget()
        self.tab_cv.setObjectName("tab_cv")
        self.tabWidget.addTab(self.tab_cv, "")
        self.horizontalLayout_18.addWidget(self.tabWidget)
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusBar = QtWidgets.QStatusBar(MainWindow)
        self.statusBar.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.statusBar.setObjectName("statusBar")
        MainWindow.setStatusBar(self.statusBar)

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "TFInterpy Gui Tool"))
        self.label_6.setText(_translate("MainWindow", "File"))
        self.pb_loadFile.setText(_translate("MainWindow", "load file"))
        self.groupBox.setTitle(_translate("MainWindow", "fields"))
        self.label.setText(_translate("MainWindow", "X"))
        self.label_2.setText(_translate("MainWindow", "Y"))
        self.label_3.setText(_translate("MainWindow", "Z"))
        self.label_4.setText(_translate("MainWindow", "P"))
        self.pb_showPoints.setText(_translate("MainWindow", "show points"))
        self.label_7.setText(_translate("MainWindow", "Variogram"))
        self.pb_calcVariogram.setText(_translate("MainWindow", "calculate variogram"))
        self.groupBox_2.setTitle(_translate("MainWindow", "mode"))
        self.rb_calcVariogramDefault.setText(_translate("MainWindow", "default"))
        self.rb_calcVariogramCustom.setText(_translate("MainWindow", "custom"))
        self.rb_calcVariogramOmni.setText(_translate("MainWindow", "omnidirectional"))
        self.label_5.setText(_translate("MainWindow", "variogram model"))
        self.gb_customVariogram.setTitle(_translate("MainWindow", "custom lags search"))
        self.label_lagInterval.setText(_translate("MainWindow", "lag interval"))
        self.label_lagNum.setText(_translate("MainWindow", "lag number"))
        self.label_lagTole.setText(_translate("MainWindow", "lag tolerent"))
        self.label_angle.setText(_translate("MainWindow", "angle"))
        self.label_angleDip.setText(_translate("MainWindow", "dip angle"))
        self.label_angleTole.setText(_translate("MainWindow", "angle tolerent"))
        self.label_bandWidth.setText(_translate("MainWindow", "band width"))
        self.le_angle.setPlaceholderText(_translate("MainWindow", "[0-PI]"))
        self.le_angleDip.setPlaceholderText(_translate("MainWindow", "[0-PI]"))
        self.label_8.setText(_translate("MainWindow", "Interpolation"))
        self.pb_executeInterp.setText(_translate("MainWindow", "execute"))
        self.gb_mode.setTitle(_translate("MainWindow", "mode"))
        self.rb_interp.setText(_translate("MainWindow", "interpolate"))
        self.rb_cv.setText(_translate("MainWindow", "cross validate"))
        self.gb_grid.setTitle(_translate("MainWindow", "grid"))
        self.label_10.setText(_translate("MainWindow", "Y"))
        self.label_14.setText(_translate("MainWindow", "Z"))
        self.label_9.setText(_translate("MainWindow", "X"))
        self.label_16.setText(_translate("MainWindow", "begin"))
        self.label_17.setText(_translate("MainWindow", "end"))
        self.label_18.setText(_translate("MainWindow", "number"))
        self.label_11.setText(_translate("MainWindow", "method"))
        self.gb_variogram.setTitle(_translate("MainWindow", "variogram"))
        self.rb_variogramNone.setText(_translate("MainWindow", "none"))
        self.rb_variogramDefault.setText(_translate("MainWindow", "default"))
        self.rb_variogramCalced.setText(_translate("MainWindow", "caculated"))
        self.label_12.setText(_translate("MainWindow", "neighborhood size"))
        self.label_alpha.setText(_translate("MainWindow", "alpha"))
        self.le_alpha.setText(_translate("MainWindow", "2"))
        self.gb_useTF.setTitle(_translate("MainWindow", "use tensorflow"))
        self.label_13.setText(_translate("MainWindow", "device"))
        self.pb_savePs.setText(_translate("MainWindow", "save points"))
        self.pb_saveGrid.setText(_translate("MainWindow", "save grid"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_interp), _translate("MainWindow", "interpolation"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_variogram), _translate("MainWindow", "variogram"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_cv), _translate("MainWindow", "cross validate"))




import matplotlib
matplotlib.use("Qt5Agg")
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.pyplot as plt

#Set the global font size of matplotlib.
plt.rcParams.update({"font.size":16})

class Canvas(FigureCanvas):
    '''
    Matploblib canvas for embedding Qt widget.
    '''

    def __init__(self, width=5, height=4, dpi=100, parent=None):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        super(Canvas, self).__init__(self.fig)
        self.setParent(parent)


# -*- coding: utf-8 -*-
# @Time    : 2022/3/25 20:05
# @Author  : UCCU

import vtk

CoolToWarm=[
    [0.0,[0,0,255]],
    [0.5,[255,255,255]],
    [1.0,[255,0,0]],
]

Rainbow=[
    [0.0,[0,0,255]],
    [0.2,[0,127,255]],
    [0.4,[0,255,0]],
    [0.6,[255,255,0]],
    [0.8,[255,165,0]],
    [1.0,[255,0,0]]
]

BlueGreenOrange=[
    [0.0,[205,228,248]],
    [0.33,[29,65,88]],
    [0.66,[221,218,153]],
    [1.0,[115,2,0]]
]

def getCTF(scalarMin,scalarMax,stops):
    '''
    Get vtk color transfer function by scalarRange and stops.

    :param scalarMin: number.
    :param scalarMax: number.
    :param stops: list, each stop represented by [percentage,[r,g,b]].
    :return: vtkDiscretizableColorTransferFunction object.
    '''
    ctf=vtk.vtkDiscretizableColorTransferFunction()
    total=scalarMax-scalarMin

    for stop in stops:
        ctf.AddRGBPoint(scalarMin+total*stop[0],stop[1][0]/255.0,stop[1][1]/255.0,stop[1][2]/255.0)
    return ctf


import pandas as pd


def readGslibPoints(filePath):
    '''
    Read points data from gslib file.

    :param filePath: str.
    :return: DataFrame object.
    '''
    columnNames = []
    with open(filePath, 'r') as fi:
        fi.readline()
        n = int(fi.readline().strip())
        for i in range(n):
            columnNames.append(fi.readline().strip())
        fi.close()
    df = pd.read_csv(filePath, sep=',|\s+', skiprows=n + 2, header=None, engine="python")
    df.columns = columnNames
    return df


def saveGslibPoints(filePath, colNames, ndarray, comment="#"):
    '''
    Save points data to gslib file.

    :param filePath: str.
    :param colNames: list, column names.
    :param ndarray: ndarray, points and properties data.
    :param comment: str, comment on the first line of the gslib file.
    :return: None.
    '''
    fo = open(filePath, 'w')
    fo.write(comment + '\n')
    fo.write('{}\n'.format(len(colNames)))
    for name in colNames:
        fo.write(name + '\n')
    df = pd.DataFrame(ndarray)
    df.to_csv(fo, sep=' ', line_terminator='\n', index=False, header=False)
    fo.close()


def saveGslibGrid(filePath, dim, begin, step, colNames, ndarray, comment=None):
    '''
    Save rectilinear grid data to gslib file.

    :param filePath: str.
    :param dim: list, [xnum, ynum, znum], xnum indicates the number of x coordinates.
    :param begin: list, [x,y,z], begin coordinates.
    :param step: list, [xstep,ystep,zstep], xstep indicates the step on x axis.
    :param colNames: list, column names.
    :param ndarray: ndarray, points and properties data.
    :param comment: str, comment on the first line of the gslib file.
    :return: None.
    '''
    fo = open(filePath, 'w')
    if comment is not None:
        fo.write('#' + comment + '\n')
    line = ' '.join(list(map(lambda x: str(x), dim)))
    fo.write(line + '\n')
    line = ' '.join(list(map(lambda x: str(x), begin)))
    fo.write(line + '\n')
    line = ' '.join(list(map(lambda x: str(x), step)))
    fo.write(line + '\n')
    df = pd.DataFrame(ndarray)
    df.columns = colNames
    df.to_csv(fo, sep=' ', line_terminator='\n', index=False)
    fo.close()



import tensorflow as tf
from tensorflow.keras import layers

tf.keras.backend.set_floatx('float64')

class LinearLayer(layers.Layer):
    '''
    The keras layer representing the linear variogram function model.
    '''

    def __init__(self, slope, C0):
        super(LinearLayer, self).__init__()
        self.slope = tf.constant(slope)
        self.C0 = tf.constant(C0)

    def call(self, h):
        return h * self.slope + self.C0


class PowerLayer(layers.Layer):
    '''
    The keras layer representing the power variogram function model.
    '''

    def __init__(self, scale, exp, C0):
        super(PowerLayer, self).__init__()
        self.scale = tf.constant(scale)
        self.exp = tf.constant(exp)
        self.C0 = tf.constant(C0)

    def call(self, h):
        return self.scale * (h ** self.exp) + self.C0


class GaussianLayer(layers.Layer):
    '''
    The keras layer representing the gaussian variogram function model.
    '''

    def __init__(self, C, a, C0):
        super(GaussianLayer, self).__init__()
        self.C = tf.constant(C)
        self.a_2 = tf.constant((a * 4 / 7) ** 2)
        self.C0 = tf.constant(C0)

    def call(self, h):
        return self.C * (1.0 - tf.exp(-1.0 * (h ** 2.0) / self.a_2)) + self.C0


class ExponentLayer(layers.Layer):
    '''
    The keras layer representing the exponent variogram function model.
    '''

    def __init__(self, C, a, C0):
        super(ExponentLayer, self).__init__()
        self.C = tf.constant(C)
        self.a_ = tf.constant(a / 3.0)
        self.C0 = tf.constant(C0)

    def call(self, h):
        return self.C * (1.0 - tf.exp(-1.0 * h / self.a_)) + self.C0


class SphericalLayer(layers.Layer):
    '''
    The keras layer representing the spherical variogram function model.
    '''

    def __init__(self, C, a, C0):
        super(SphericalLayer, self).__init__()
        self.C = tf.constant(C)
        self.a = tf.constant(a)
        self.a2 = tf.constant(2 * a)
        self.a2_3 = tf.constant(2 * self.a ** 3)
        self.C0 = tf.constant(C0)

    def call(self, h):
        condition = tf.greater(h, self.a)
        return tf.where(condition, self.C + self.C0, self.C * (h * 3.0 / self.a2 - h ** 3.0 / self.a2_3) + self.C0)


class NestVariogramLayer(layers.Layer):
    '''
    The keras layer representing the nested variogram functions.
    '''

    def __init__(self, variogramLayers, unitVectors):
        '''
        Construct a nested variogram function layer.

        :param variogramLayers: list, containing variogram layers.
        :param unitVectors: array_like, unit vectors corresponding to the direction.
        '''
        super(NestVariogramLayer, self).__init__()
        self.variogramLayers = variogramLayers
        self.unitVectors = unitVectors
        if len(self.unitVectors.shape) == 2:
            self.unitVectors = self.unitVectors.reshape((*self.unitVectors.shape, 1))

    def call(self, vec):
        total = 0
        for idx, variogramLayer in enumerate(self.variogramLayers):
            h = tf.abs(tf.matmul(vec, self.unitVectors[idx]))
            var = variogramLayer(h)
            total += var
        return total


VariogramLayerMap = {
    "gaussian": GaussianLayer,
    "exponent": ExponentLayer,
    "spherical": SphericalLayer,
    "linear": LinearLayer,
    "power": PowerLayer,
}


def getVariogramLayer(variogramBuilder):
    '''
    Construct variogram layer by variogram builder.

    :param variogramBuilder: VariogramBuilder object.
    :return: keras' layer object.
    '''
    LayerClass = VariogramLayerMap[variogramBuilder.model]
    vl = LayerClass(*variogramBuilder.params)
    return vl


def getNestVariogramLayer(variogramBuilders, unitVectors):
    '''
    Construct nested variogram layer by variogram builders and corresponding unit vectors.

    :param variogramBuilders: list, containing variogram builders.
    :param unitVectors: array_like, unit vectors corresponding to the direction.
    :return: keras' layer object.
    '''
    vls = []
    for vb in variogramBuilders:
        vl = getVariogramLayer(vb)
        vls.append(vl)
    return NestVariogramLayer(vls, unitVectors)



from tfinterpy.utils import *
from tfinterpy.variogramExp import search2d, search3d
from tfinterpy.variogramModel import variogramModel, VariogramModelMap
from scipy.optimize import least_squares
from functools import partial

EPS = 1e-16

def resident(params, x, y, variogram_function):
    '''
    This function is used to calculate resident.

    :param params: tuple, represents the parameters passed to the variogram function.
    :param x: number or ndarray, distance.
    :param y: number or ndarray, semivariance.
    :param variogram_function: function, variogram function.
    :return: number or ndarray, resident=( variogram(x,params[0],...,params[n])-y )^2
    '''
    # error = variogram_function(x, *params) - y
    error = variogram_function(*params, x) - y
    return error ** 2


def getX0AndBnds(h, y, variogram_model):
    '''
    Calculate the initial value and range of parameters(nugget, sill, range, ...)
    according to the variogram function model.

    :param h: array_like, distance.
    :param y: array_like, semivariance.
    :param variogram_model: str, indicates the variogram function model.
    :return: tuple, (x0,bnds). x0 represents the initial value of the parameters,
        and bnds represents the range of the parameters.
    '''
    if variogram_model == "linear":
        x0 = [(np.amax(y) - np.amin(y)) / 2 / (np.amax(h) - np.amin(h)), np.amin(y)]
        bnds = ([0.0, 0.0], [np.inf, np.amax(y)])
    elif variogram_model == "power":
        x0 = [(np.amax(y) - np.amin(y)) / (np.amax(h) - np.amin(h)), 1.0, np.amin(y)]
        bnds = ([0.0, 0.001, 0.0], [np.inf, 1.999, np.amax(y)])
    else:
        x0 = [(np.amax(y) - np.amin(y)) / 2, 0.5 * np.amax(h), np.amin(y)]
        bnds = ([0.0, 0.0, 0.0], [np.amax(y), np.amax(h), np.amax(y)])
    return x0, bnds


class VariogramBuilder:
    '''
    VariogramBuilder class is used to build variogram function.
    '''

    def __init__(self, lags, model="spherical"):
        '''
        The VariogramBuilder is constructed with lags and string indicating the model.

        :param lags: array_like, a lag is [distance,semivariance].
        :param model: str, indicates the variogram function model.
        '''
        self.model = model
        self.lags = np.array(lags)
        x0, bnds = getX0AndBnds(self.lags[:, 0], self.lags[:, 1], model)
        res = least_squares(resident, x0, bounds=bnds, loss="huber",
                            args=(self.lags[:, 0], self.lags[:, 1], variogramModel(model)))
        self.params = res.x
        self.mae = np.mean(res.fun)

    def showVariogram(self, axes=None):
        '''
        Plot variogram.

        :param axes: axes, if axes is set to None, plot on a new axes.
        :return: None.
        '''
        if axes is None:
            import matplotlib.pyplot as axes
        variogram = self.getVariogram()
        axes.scatter(self.lags[:, 0], self.lags[:, 1], alpha=0.5)
        max = np.max(self.lags[:, 0])
        X = np.arange(0, max, max / 100)
        Y = variogram(X)
        axes.plot(X, Y, alpha=0.5, color="red", label=self.model)

    def getVariogram(self):
        '''
        Get variogram function(closure function).
        :return: function.
        '''
        func=variogramModel(self.model)
        pfunc=partial(func,*self.params)
        return pfunc


class NestVariogram:
    '''
    NestVariogram class is used to calculate the nesting variogram function value.
    A NestVariogram object is a functor.
    '''

    def __init__(self, variograms, unitVectors):
        '''
        The nested variational function object is constructed by
        using a number of unit vectors and variograms calculated
        from different directions.

        :param variograms: list, variogram functions.
        :param unitVectors: array_like, unit vectors corresponding to the direction.
        '''
        self.variograms = variograms
        self.unitVectors = unitVectors

    def __call__(self, vec):
        '''
        Calculate the value of the nesting variogram function.

        :param vec: array_like, vectors.
        :return: array_like, variogram function value.
        '''
        totalVar = 0
        for idx, unitVector in enumerate(self.unitVectors):
            h = np.abs(np.dot(vec, unitVector))
            var = self.variograms[idx](h)
            totalVar += var
        return totalVar


def calculateOmnidirectionalVariogram2D(samples, partitionNum=8, leastPairNum=10, lagNum=20, lagInterval=None,
                                        lagTole=None, bandWidth=None, model=None):
    '''
    Calculating the omnidirectional variogram in two dimensions.

    :param samples: ndarray, array containing all sample points. The last column must be the properties.
        Each item is represented by [x,y,property].

    :param partitionNum: integer, indicates how many parts the angle range is divided into.
    :param leastPairNum: integer, at least lastPairNum lags need to be searched
        in one direction to fit the variogram function.
    :param lagNum: integer, number of lags to search.
    :param lagInterval: number, interval lags' distance.
        If lagInterval is set to None, then it is set to 0.75 times.
        the maximum vector modulus length divided by lagNum.
    :param lagTole: number, tolerance of lag's distance.
        If lagTole is set to None, then it is set to lagInterval/2.
    :param bandWidth: number, bandwidth used during search.
        If bandWidth is set to None, then it is set to the mean
        of the vector modulus length divided by 2.
    :param model: str, specify which variogram function model to use for the fit.
        If model is set to None, fitting will be attempted using all the variogram
        function models in the variogramModel module, and the best fit will be used.
    :return: tuple, (NestVariogram,variogramBuilders).
        NestVariogram is NestVariogram object,
        variogramBuilders is a list of VariogramBuilder objects corresponding to
        all variogram functions maintained within the NestVariogram object.
    '''
    vecs = calcVecs(samples, repeat=False)
    vars = 0.5 * vecs[:, 2] ** 2
    vecs = vecs[:, :2]
    azimuthStep = np.pi / partitionNum
    unitVectors = []
    azimuths = []
    for i in range(partitionNum):
        azimuth = (i + 0.5) * azimuthStep
        azimuths.append(azimuth)
        unitVectors.append([np.cos(azimuth), np.sin(azimuth)])
    unitVectors = np.array(unitVectors)
    azimuths = np.array(azimuths)
    norms = np.linalg.norm(vecs, axis=1) + EPS
    if bandWidth is None:
        bandWidth = norms.mean() / 2
    if lagInterval is None:
        lagInterval = norms.max() * 0.75 / lagNum
    if lagTole is None:
        lagTole = lagInterval / 2
    lagRanList = [[(i + 1) * lagInterval - lagTole, (i + 1) * lagInterval + lagTole] for i in range(lagNum)]
    bucket = [[[] for j in range(lagNum)] for i in range(partitionNum)]
    thetas = np.arctan2(vecs[:, 1], vecs[:, 0])
    thetas[thetas < 0] += np.pi
    indiceTheta = thetas // azimuthStep
    indiceTheta = indiceTheta.astype('int')
    for i in range(partitionNum):
        deltas = thetas - azimuths[i]
        bands = norms * deltas
        for j in range(lagNum):
            indice = np.where(
                ((indiceTheta == i) & (norms > lagRanList[j][0]) & (norms < lagRanList[j][1]) & (bands < bandWidth)))[0]
            bucket[i][j] = list(zip(norms[indice], vars[indice]))
    processedBucket = [[] for i in range(partitionNum)]
    searchedPairNumRecords = [[] for i in range(partitionNum)]
    for i in range(partitionNum):
        for j in range(lagNum):
            searchedPairNum = len(bucket[i][j])
            if searchedPairNum < 3:
                continue
            searchedPairNumRecords[i].append(searchedPairNum)
            lag = np.mean(bucket[i][j], axis=0)
            processedBucket[i].append(lag)
    lagsList = [None for i in range(partitionNum)]
    availableDir = []
    for i in range(partitionNum):
        if len(processedBucket[i]) > leastPairNum:
            lagsList[i] = processedBucket[i]
            availableDir.append(i)
    variogramBuilders = []
    if model is None:
        for lags in lagsList:
            if lags is None:
                continue
            minResident = float('+inf')
            best = None
            for key in VariogramModelMap:
                vb = VariogramBuilder(lags, key)
                if vb.mae < minResident:
                    minResident = vb.mae
                    best = vb
            variogramBuilders.append(best)
    else:
        for lags in lagsList:
            if lags is None:
                continue
            vb = VariogramBuilder(lags, model)
            variogramBuilders.append(vb)
    nestVariogram = NestVariogram([vb.getVariogram() for vb in variogramBuilders], unitVectors[availableDir])
    return nestVariogram, variogramBuilders


def calculateOmnidirectionalVariogram3D(samples, partitionNum=[6, 6], leastPairNum=10, lagNum=20, lagInterval=None,
                                        lagTole=None, bandWidth=None, model=None):
    '''
    Calculating the omnidirectional variogram in three dimensions.

    :param samples: ndarray, array containing all sample points. The last column must be the properties.
        Each item is represented by [x,y,z,property].
    :param partitionNum: array_like,Indicates how many parts the angle range
        in horizontal and vertical directions is divided into.
    :param leastPairNum: integer, at least lastPairNum lags need to be searched
        in one direction to fit the variogram function.
    :param lagNum: integer, number of lags to search.
    :param lagInterval: number, interval lags' distance.
        If lagInterval is set to None, then it is set to 0.75 times.
    :param lagTole: number, tolerance of lag's distance.
        If lagTole is set to None, then it is set to lagInterval/2.
    :param bandWidth: number, bandwidth used during search.
        If bandWidth is set to None, then it is set to the mean
        of the vector modulus length divided by 2.
    :param model: str, specify which variogram function model to use for the fit.
        If model is set to None, fitting will be attempted using all the variogram
        function models in the variogramModel module, and the best fit will be used.
    :return: tuple, (NestVariogram,variogramBuilders).
        NestVariogram is NestVariogram object,
        variogramBuilders is a list of VariogramBuilder objects corresponding to
        all variogram functions maintained within the NestVariogram object.
    '''
    vecs = calcVecs(samples, repeat=False)
    vars = 0.5 * vecs[:, 3] ** 2
    vecs = vecs[:, :3]
    if type(partitionNum) != list:
        partitionNum = [partitionNum, partitionNum]
    azimuthStep = np.pi / partitionNum[0]
    azimuths = [(i + 0.5) * azimuthStep for i in range(partitionNum[0])]
    dipStep = np.pi / partitionNum[1]
    dips = [(i + 0.5) * dipStep for i in range(partitionNum[1])]
    unitVectors = []
    for i in range(partitionNum[0]):
        a = azimuths[i]
        for j in range(partitionNum[1]):
            b = dips[j]
            unitVectors.append([np.cos(a) * np.cos(b), np.sin(a) * np.cos(b), np.sin(b)])
    unitVectors = np.array(unitVectors)
    azimuths = np.array(azimuths)
    dips = np.array(dips)
    norms = np.linalg.norm(vecs, axis=1) + EPS
    if bandWidth is None:
        bandWidth = norms.mean() / 2
    if lagInterval is None:
        lagInterval = norms.max() * 0.75 / lagNum
    if lagTole is None:
        lagTole = lagInterval / 2
    lagRanList = [[(i + 1) * lagInterval - lagTole, (i + 1) * lagInterval + lagTole] for i in range(lagNum)]
    bucket = [[[[] for k in range(lagNum)] for j in range(partitionNum[1])] for i in range(partitionNum[0])]
    thetas1 = np.arctan2(vecs[:, 1], vecs[:, 0])
    thetas1[thetas1 < 0] += np.pi
    thetas2 = np.arctan2(vecs[:, 2], np.linalg.norm(vecs[:, :2], axis=1))
    thetas2[thetas2 < 0] += np.pi
    indiceTheta1 = thetas1 // azimuthStep
    indiceTheta1 = indiceTheta1.astype('int')
    indiceTheta2 = thetas2 // dipStep
    indiceTheta2 = indiceTheta2.astype('int')
    for i in range(partitionNum[0]):
        for j in range(partitionNum[1]):
            unitVector = unitVectors[i * partitionNum[1] + j]
            angles = np.arccos(np.dot(vecs, unitVector) / norms)
            indice = np.where(angles > np.pi / 2)[0]
            angles[indice] = np.pi - angles[indice]
            bands = norms * np.sin(angles)
            for k in range(lagNum):
                indice = np.where(
                    ((indiceTheta1 == i) & (indiceTheta2 == j) & (norms > lagRanList[k][0]) & (
                            norms < lagRanList[k][1]) & (bands < bandWidth)))[0]
                bucket[i][j][k] = list(zip(norms[indice], vars[indice]))
    processedBucket = [[[] for j in range(partitionNum[1])] for i in range(partitionNum[0])]
    searchedPairNumRecords = [[[] for j in range(partitionNum[1])] for i in range(partitionNum[0])]
    for i in range(partitionNum[0]):
        for j in range(partitionNum[1]):
            for k in range(lagNum):
                searchedPairNum = len(bucket[i][j][k])
                if searchedPairNum < 3:
                    continue
                searchedPairNumRecords[i][j].append(searchedPairNum)
                lag = np.mean(bucket[i][j][k], axis=0)
                processedBucket[i][j].append(lag)
    lagsList = [None for i in range(partitionNum[0] * partitionNum[1])]
    availableDir = []
    for i in range(partitionNum[0]):
        for j in range(partitionNum[1]):
            if len(processedBucket[i][j]) > leastPairNum:
                idx = i * partitionNum[1] + j
                lagsList[idx] = processedBucket[i][j]
                availableDir.append(idx)
    variogramBuilders = []
    if model is None:
        for lags in lagsList:
            if lags is None:
                continue
            minResident = float('+inf')
            best = None
            for key in VariogramModelMap:
                vb = VariogramBuilder(lags, key)
                if vb.mae < minResident:
                    minResident = vb.mae
                    best = vb
            variogramBuilders.append(best)
    else:
        for lags in lagsList:
            if lags is None:
                continue
            vb = VariogramBuilder(lags, model)
            variogramBuilders.append(vb)
    nestVariogram = NestVariogram([vb.getVariogram() for vb in variogramBuilders], unitVectors[availableDir])
    return nestVariogram, variogramBuilders


def calculateDefaultVariogram2D(samples, model='spherical'):
    '''
    Calculate the default variogram function in two dimensions,
    without considering the effect of direction.

    :param samples: ndarray, array containing all sample points. The last column must be the properties.
        Each item is represented by [x,y,property].
    :param model: str, specify which variogram function model to use for the fit.
    :return: VariogramBuilder object.
    '''
    vecs = calcVecs(samples, repeat=False)
    hav = calcHAVByVecs(vecs)
    lagNum = 20
    lagInterval = hav[:, 0].max() / 2 / lagNum
    lagTole = lagInterval * 0.5
    angles = hav[:, 1]
    buckets = []
    portionNum = 8
    portion = np.pi / portionNum
    for i in range(portionNum):
        count = np.where((angles > portion * i) & (angles < portion * (i + 1)))[0].shape[0]
        buckets.append(count)
    idx = np.argmax(buckets)
    angle = portion * idx
    angleTole = portion
    bandWidth = hav[:, 0].mean() / 2
    lags, _ = search2d(vecs[:, :2], hav[:, 2], lagNum, lagInterval, lagTole, angle, angleTole, bandWidth)
    if len(lags) < 3:
        lags = calcHV(samples)
    vb = VariogramBuilder(lags, model)
    return vb


def calculateDefaultVariogram3D(samples, model='spherical'):
    '''
    Calculate the default variogram function in three dimensions,
    without considering the effect of direction.

    :param samples: ndarray, array containing all sample points. The last column must be the properties.
        Each item is represented by [x,y,z,property].
    :param model: str, specify which variogram function model to use for the fit.
    :return: VariogramBuilder object.
    '''
    vecs = calcVecs(samples, repeat=False)
    habv = calcHABVByVecs(vecs)
    lagNum = 20
    lagInterval = habv[:, 0].max() / 2 / lagNum
    lagTole = lagInterval * 0.5
    azimuths = habv[:, 1]
    dips = habv[:, 2]
    buckets = []
    portionNum = 5
    portion = np.pi / portionNum
    for i in range(portionNum):
        for j in range(portionNum):
            count = np.where((azimuths > portion * i) & (azimuths < portion * (i + 1)) & (dips > portion * j) & (
                    dips < portion * (j + 1)))[0].shape[0]
            buckets.append(count)
    idx = np.argmax(buckets)
    azimuth = idx // portionNum * portion
    dip = idx % portionNum * portion
    angleTole = portion / 2
    bandWidth = habv[:, 0].mean() / 2
    lags, _ = search3d(vecs[:, :3], habv[:, 3], lagNum, lagInterval, lagTole, azimuth, dip, angleTole, bandWidth)
    if len(lags) < 3:
        lags = calcHV(samples)
    vb = VariogramBuilder(lags, model)
    return vb



from tfinterpy.idw import IDW
from tfinterpy.grid import Grid2D
from examples.ncFileUtil import getSamples
from examples.plotUtils import *
import numpy as np
import random

if __name__ == "__main__":
    random.seed(1)
    W, H = 100, 100
    M = 100
    offset = (400, 800)
    N = 8

    # Get sample points and original elevation data from netcdf file
    # containing Digital Elevation Model data.
    samples, lats, lons, ele = getSamples("data/tatitlek_815_mhhw_2011.nc", 'lat', 'lon', 'Band1',
                                          offset, (W, H), M)
    # Create linear 2D grid.
    grid = Grid2D()
    grid.rectlinear((W, H), (1, W), (1, H))

    exe = IDW(samples)# Create a idw interpolator.
    grid.pro = exe.execute(grid.points(), N)# Perform interpolation of all points in the grid.

    print(exe.crossValidate(N))# Perform leave-one-out validation and print result.
    print(exe.crossValidateKFold(10, N))# Perform k-fold validation and print result.

    pro = grid.pro.reshape((grid.dim[1], grid.dim[0]))# Reshape properties as 2d ndarray.
    plt.figure()
    plt.subplot(221)
    img2d(ele, title="Origin")# Plotting original elevation data.
    plt.subplot(222)
    img2d(pro, title="IDW")# Plotting interpolation results.
    absdiff = np.abs(pro - ele)
    plt.subplot(223)
    img2d(absdiff, title="AE")# Plotting absolute errors.
    print('result MAE:', np.mean(absdiff))
    plt.show()
